#include "xwio.h"
#include <xview/textsw.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>

/* Trace Status Flags */
#define T_S_WAITING     1
#define T_S_ACTIVE      2
#define T_S_AUTO        3
#define T_S_RUNNING     4
#define T_S_TERMINATED  5
#define T_S_PRINTING    6


struct DebugWindowStruct {
	Frame	frame;
	Panel_item	message;
	struct TwinStruct *twin;
	Textsw	textsw;
	int	outwin;
	int	ignore;
	};

extern struct pidref *object_plist;
extern struct list *trace_list;

struct TraceDataStruct {    /* List of TraceProcesses */
	int	Tid;
	long	Pid;
	int 	RTid;
	long	RPid;
	int	nStatus;
	int	IgnoreTrace;
	int	nListPos;
	struct DebugWindowStruct *DebugWindow;
	int	lFilePos;
	char	FileName[T_NAMESIZE];
	char	PredName[T_NAMESIZE];
	char	Display[T_DISPLAY_X*T_DISPLAY_Y];
	char	ProcName[PROC_NAME_SIZE];
	int	start_y;
	int	cur_x;
	int	cur_y;
	};

struct 	{
	Frame		frame;
	Frame		parent;
	Panel_item	list;
	Panel_item	choice;
	int		ignore;
	} trace_dlg;

char	*StateName[]={
	"Zero ? ",
	"waiting",
	"active ",
	"auto   ",
	"running",
	"terminated",
	"printing"};

extern	int	fRunAll;
extern struct ConsoleStruct console;
extern struct TwinStruct *dummy_twin;

void trace_list_set (struct TraceDataStruct *p);
void debug_win_set (struct TraceDataStruct *p);
void debug_win_print (struct TraceDataStruct *p, char *text);
void debug_win_reset (struct DebugWindowStruct *dbw);
void debug_show (struct TraceDataStruct *p);
void debug_close (struct DebugWindowStruct *dbw);
void debug_open (struct TraceDataStruct *p);

void debug_step (struct TraceDataStruct *p)
/*
 *	send return for trace msg
 */
	{
	switch (p->nStatus) 
		{
		case T_S_AUTO:
			p->nStatus=T_S_ACTIVE;
			break;
		case T_S_WAITING:
			SendRet (p->RTid, p->RPid, 1);
			p->nStatus=T_S_ACTIVE;
			break;
		case T_S_RUNNING:
			{
			long	msg[4];
			msg[0]=5;
			msg[1]=p->Pid;
			msg[2]=1;
			SendSysMessage (p->Tid, 0, 3, msg);
			p->nStatus=T_S_ACTIVE;
			break;
			}
		}
	trace_list_set (p);
	debug_win_set (p);
	}

void debug_set_auto (struct TraceDataStruct *p)
/*
 *	switch process to auto mode
 */
	{
	switch (p->nStatus)
		{
		case T_S_WAITING:
			SendRet (p->RTid, p->RPid, 1);
			break;
		case T_S_RUNNING:
			{
			long	msg[4];
			msg[0]=5;
			msg[1]=p->Pid;
			msg[2]=1;
			SendSysMessage (p->Tid, 0, 3, msg);
			break;
			}
		}
	p->nStatus=T_S_AUTO;
	trace_list_set (p);
	debug_win_set (p);
	}

void debug_set_run (struct TraceDataStruct *p)
/*
 *	switch process to running mode...
 */
	{
	if (p->nStatus==T_S_WAITING)
		SendRet (p->RTid, p->RPid, 0L);
	p->nStatus=T_S_RUNNING;
	trace_list_set (p);
	debug_win_set (p);
	}

void trace_list_add (struct TraceDataStruct *p)
/* 
 *	Add entry to the trace list
 */
	{
	char	s[100];

	strcpy (s,p->FileName);
	if (p->nStatus==T_S_WAITING)
		sprintf (s,"   %6d/%-6d %-15s : %s  %s (%d)",p->Tid, p->Pid, p->ProcName, StateName[p->nStatus],p->FileName, p->lFilePos);
	else
		sprintf (s,"   %6d/%-6d %-15s : %s",p->Tid, p->Pid, p->ProcName, StateName[p->nStatus]);
	if (p->DebugWindow)
		s[0]='*';
	if (p->IgnoreTrace)
		s[1]='I';
	xv_set (trace_dlg.list,
		PANEL_LIST_INSERT,	0,
		PANEL_LIST_STRING,	0,	s,
		PANEL_LIST_CLIENT_DATA, 0,	p,
		PANEL_LIST_DISPLAY_ROWS,	10,
		NULL);
	}

int trace_list_get_row (struct TraceDataStruct *p)
/*
 *	Find the row where the process is currently displayed
 */
	{
	int	i;
	struct TraceDataStruct *q;

	i=(int) xv_get (trace_dlg.list, PANEL_LIST_NROWS);

	while (i--)
		{
		q=(struct TraceDataStruct *)xv_get (trace_dlg.list, PANEL_LIST_CLIENT_DATA, i);
		if (q==p)
			return i;
		}
	return -1;
	}

void trace_list_set (struct TraceDataStruct *p)
/*
 *	update a process entry in the trace list
 */
	{
	int i;

	if (!trace_dlg.frame)
		return;
	i=trace_list_get_row (p);
	if (i != -1)
		{
		char s[100];
	
		if (p->nStatus==T_S_WAITING)
			sprintf (s,"  %6d/%-6d %-15s : %s  %s (%d)",p->Tid, p->Pid, p->ProcName, StateName[p->nStatus],p->FileName, p->lFilePos);
		else
			sprintf (s,"  %6d/%-6d %-15s : %s",p->Tid, p->Pid, p->ProcName, StateName[p->nStatus]);

		if (p->DebugWindow)
			s[0]='*';
		if (p->IgnoreTrace)
			s[1]='I';

		xv_set (trace_dlg.list, 
			PANEL_LIST_STRING,	i, s,
			NULL);
		}
	else
		trace_list_add (p);
	}

void trace_list_del (struct TraceDataStruct *p)
	{
	int i;

	if (!trace_dlg.frame)
		return;
	i=trace_list_get_row (p);
	if (i != -1)
		xv_set (trace_dlg.list, PANEL_LIST_DELETE, i, NULL);
        }

struct TraceDataStruct *trace_list_selected (void)
/*
 *	return the data struct for the currently selected process
 */
	{
	int	s;

	s=(int) xv_get (trace_dlg.list, PANEL_LIST_FIRST_SELECTED);
	if (s==-1)
		return NULL;
	return (struct TraceDataStruct *)xv_get (trace_dlg.list, PANEL_LIST_CLIENT_DATA, s);
	}

void trace_proc_status (Panel_item item, int val, Event *event)
/*
 *	choice item notify function:
 *	the current process status has been changed from the trace_dlg
 */
	{
	struct TraceDataStruct *p;

	p=trace_list_selected ();	/* get the current selection */
	if (!p)
		return;
	switch (val)
		{
		case 0:
			debug_step (p);
			break;
		case 1:
			debug_set_auto (p);
			break;
		case 2:
			debug_set_run (p);
			break;
		}
	}

void trace_proc_add (Panel_item item, Event *event)
/*
 *	"Add" button notify function:
 *	add a process from process list to tracelist...
 */
	{
	process_list (trace_dlg.frame, 1);
	}

void trace_proc_remove (Panel_item item, Event *event)
/*
 *	"Remove" button notify function:
 *	remove the current entry from the list
 */
	{
	struct TraceDataStruct *p;

	p=trace_list_selected ();	/* get the current selection */
	if (!p)
		return;
	debug_set_run (p);
	trace_list_del (p);
	UnTrace (p);
	}

void trace_proc_debug (Panel_item item, Event *event)
/*
 *	"Debug" button notify function:
 *	open a debug window for currently selected process
 */
	{
	struct TraceDataStruct *p;

	p=trace_list_selected ();
	if (!p)
		return;
	if (p->DebugWindow)
		{
		if (xv_get ((p->DebugWindow)->frame, FRAME_CLOSED))
			xv_set ((p->DebugWindow)->frame, FRAME_CLOSED, FALSE, NULL);
		xv_set ((p->DebugWindow)->frame, XV_SHOW, TRUE, NULL);
		}
	else	
		debug_open (p);
	}

int trace_proc_select (Panel_item item, char *string, Xv_opaque client_data, Panel_list_op op, int row)
/*
 *	TraceList notify function:
 *	Selection has been made in trace list, update status choice...
 */
	{
	struct TraceDataStruct *p;

	p=(struct TraceDataStruct *)client_data;

	if (op==PANEL_LIST_OP_SELECT)
		if (p->nStatus==T_S_RUNNING)
			xv_set (trace_dlg.choice, PANEL_VALUE, 2, NULL);
		else
			if (p->nStatus==T_S_AUTO)
				xv_set (trace_dlg.choice, PANEL_VALUE, 1, NULL);
			else
				xv_set (trace_dlg.choice, PANEL_VALUE, 0, NULL);
	return XV_OK;
	}

void trace_proc_ignoretrace (Panel_item item, int value, Event *event)
/*
 *	"Ignore Trace" notify function:
 *	set the ignore trace flag
 */
	{
	if (value)
		fRunAll=TRUE;
	else
		fRunAll=FALSE;
	}

void trace_proc_ok (Panel_item item, Event *event)
	{
	xv_destroy_safe (trace_dlg.frame);
	trace_dlg.frame=0;
	}

void trace_done_proc (Frame frame)
/*
 *	something has closed the TraceList...
 */
	{
	trace_proc_ok (NULL, NULL);
	}

void trace_window (Frame parentframe)
/*
 *	display the tracelist dialog...
 */
	{
	struct list *elem;
	struct TraceDataStruct *p;
	Panel	panel;

	if (trace_dlg.frame)
		return;
	trace_dlg.parent=parentframe;
	trace_dlg.frame= (Frame) xv_create (parentframe, FRAME,
		FRAME_LABEL,		"Trace list",
		WIN_COLUMNS,		60,
		WIN_ROWS,		22,
		FRAME_DONE_PROC,	trace_done_proc,
		NULL);
	panel= (Panel) xv_create (trace_dlg.frame, PANEL,  NULL);
	trace_dlg.choice= (Panel_item)xv_create (panel, PANEL_CHOICE,
		PANEL_LAYOUT,		PANEL_VERTICAL,
		PANEL_CHOICE_STRINGS,	"Wait", "Auto", "Run", NULL,
		PANEL_NOTIFY_PROC,	trace_proc_status,
		NULL);
	trace_dlg.list=(Panel_item)xv_create (panel, PANEL_LIST,
		PANEL_LAYOUT,		PANEL_VERTICAL,
		PANEL_NEXT_COL,		-1,
		PANEL_LIST_WIDTH,	-1,
		PANEL_LIST_DISPLAY_ROWS,10,
		PANEL_NOTIFY_PROC,	trace_proc_select,
		NULL);
	trace_dlg.ignore= (Panel_item)xv_create (panel, PANEL_CHECK_BOX,
		PANEL_LAYOUT,		PANEL_HORIZONTAL,
		PANEL_NEXT_ROW,		-1,
		PANEL_CHOICE_STRINGS,	"Ignore Trace", NULL,
		PANEL_NOTIFY_PROC,	trace_proc_ignoretrace,
		NULL);
	if (fRunAll)
		xv_set (trace_dlg.ignore, PANEL_VALUE, 1, NULL);
	else
		xv_set (trace_dlg.ignore, PANEL_VALUE, 0, NULL);

	xv_create (panel, PANEL_BUTTON,
		PANEL_LAYOUT,		PANEL_HORIZONTAL,
		PANEL_NEXT_ROW,		-1,
		PANEL_NOTIFY_PROC,	trace_proc_ok,
		PANEL_LABEL_STRING,	"Ok",
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_NOTIFY_PROC,	trace_proc_add,
		PANEL_LABEL_STRING,	"Add",
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_NOTIFY_PROC,	trace_proc_debug,
		PANEL_LABEL_STRING,	"Debug",
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_NOTIFY_PROC,	trace_proc_remove,
		PANEL_LABEL_STRING,	"Remove",
		NULL);

	elem=get_first_list_elem (trace_list);
	while (elem)
		{
		p=(struct TraceDataStruct *)elem->data;
		trace_list_add (p);
		elem=get_next_list_elem (elem);
		}
	window_fit (trace_dlg.frame);
	xv_set (trace_dlg.frame, XV_SHOW, TRUE, NULL);
        }

struct TraceDataStruct *FindTraceData (int Tid, long Pid)
/* returns Pointer to TraceDataStruct */
	{
	struct list *elem;
	struct TraceDataStruct *p;

	elem=get_first_list_elem (trace_list);
	while (elem)
		{
		p=(struct TraceDataStruct *)elem->data;
		if (p->Tid==Tid && p->Pid==Pid)
			return p;  /* Process found */
		elem=get_next_list_elem (elem);
		}
	return NULL;
        }

int TraceOn (int Tid, long Pid)
/* add Traced Process to List if not there already, set correct Status
   and notify Debug-Windows or TraceDlg... */

	{
	struct ProcDataStruct *proc;
	struct TraceDataStruct *p;
	struct list *elem;

	if (!(p=FindTraceData (Tid, Pid)))
		{
		p=(struct TraceDataStruct *)malloc (sizeof (struct TraceDataStruct));
		elem=(struct list *)malloc (sizeof (struct list));
		elem->data=p;
		assertz_list (trace_list, elem);
		p->Tid=Tid;
		p->Pid=Pid;
		p->DebugWindow=0;
		p->nListPos=-1;
		p->lFilePos=0;
		memset (p->FileName, '\0', T_NAMESIZE);
		memset (p->PredName, '\0', T_NAMESIZE);
		memset (p->Display, ' ', T_DISPLAY_X*T_DISPLAY_Y);

		proc=FindProcData (Tid, Pid);
		if (proc!=NULL)
			strcpy (p->ProcName, proc->ProcName);
		else
			strcpy (p->ProcName, "");
		p->cur_x=0;
		p->cur_y=0;
		p->start_y=0;
		p->nStatus = T_S_RUNNING;
		p->IgnoreTrace = FALSE;
		}

	if (!((fRunAll || p->IgnoreTrace) && p->nStatus==T_S_RUNNING))
		p->nStatus=T_S_ACTIVE;

	trace_list_set (p);
	debug_win_set (p);

	if (p->nStatus==T_S_ACTIVE)
		if (p->DebugWindow)
			debug_show (p);
		else
			if (!trace_dlg.frame)
				trace_window (console.frame);
			else
				xv_set (trace_dlg.frame, XV_SHOW, TRUE, NULL);
	if (p->nStatus==T_S_ACTIVE)
		return 1;
	else
		return 0;
	}

int TraceOff (int Tid, int Pid)
/* Set TraceStatus of Process to Running if not overridden
   and notify Debug-Windows or TraceDlg... */

	{
	int	i;
	struct TraceDataStruct *p;

	if (!(p=FindTraceData (Tid, Pid)))
		return 0;	/* process not found, consider it running... */

	if (p->nStatus==T_S_RUNNING)
		return 0;

	trace_list_set (p);
	debug_win_set (p);

	if (p->nStatus==T_S_ACTIVE)
		return 1;	/* return value is usually 0, TraceDialog or */
	else			/* DebugWindow can override TraceOff and return 1*/
		return 0;
	}

int Trace (int RTid, int RPid, int Tid, int Pid, int FilePos, char *FileName, char *Text)
/* Process has stepped. Set Status to waiting and notify Trace List or
   Debug Window or (if auto) return Step Command (1) immediatly... */

	{
	struct TraceDataStruct *p;
	struct ProcDataStruct *proc;

	if (!(p=FindTraceData (Tid, Pid)))
		{
		if (0==TraceOn (Tid, Pid))      /* Simulate Trace On */
			return 0;		/* fRunAll is set */
		if (!(p=FindTraceData (Tid, Pid)))
			return 0;       /* give up */
		}

	if (p->nStatus==T_S_RUNNING)
		return 0;	/* Switch off Trace */

	if (p->nStatus==T_S_TERMINATED)
		{				/* Process seems UNDEAD ... */
		p->nStatus=T_S_RUNNING;
		trace_list_set (p);
		return 0;			/* I don't believe in ghosts */
		}

	p->RTid=RTid;
	p->RPid=RPid;
	strncpy(p->FileName, FileName, T_NAMESIZE);
	p->FileName[T_NAMESIZE-1]='\0';
	p->lFilePos=FilePos;
	proc=FindProcData (Tid, Pid);
	if (proc!=NULL)
		strcpy (p->ProcName, proc->ProcName);
	else	
		strcpy (p->ProcName, "");

	if (p->nStatus==T_S_AUTO)
		{
		trace_list_set (p);
		debug_win_set (p);
		debug_win_print (p, Text);
		return 1;	/* Simulate continue */
		}
	p->nStatus=T_S_WAITING;
	trace_list_set (p);
	debug_win_set (p);
	debug_win_print (p,Text);
	return -1;		/* Don't send RetMsg */
	}

int TracePrint (int Tid, int Pid, char *Text)
/* Print additional Information into Display WIndow. */
	{
	int	i;
	struct TraceDataStruct *p;

	if (!(p=FindTraceData (Tid, Pid)))
		return 0;
	if (p->DebugWindow)
		debug_win_print (p, Text);
	else
		if (p->nStatus==T_S_RUNNING)
			return 0;
	return 1;
	}

int UnTrace (struct TraceDataStruct *p)
/* remove Process from TraceList, called from TraceDialog or DebugWindow */
	{
	struct list *elem;

	if (p->nStatus==T_S_WAITING)
		SendRet (p->RTid, p->RPid, 0L);
	if (p->DebugWindow)
		debug_win_reset (p->DebugWindow);

	elem=find_list_ptr (trace_list, p);
	if (! elem)
		return 0;
	remove_list_elem (trace_list, elem);
	free (elem);
	free (p);
        
	return 1;
	}

void TraceSave (struct TraceDataStruct *p, struct TwinStruct *twin)
/* Copy Trace output from hWnd to TraceListData */
	{
	memcpy (p->Display, twin->screen, T_DISPLAY_X*T_DISPLAY_Y);
	p->cur_x=twin->cur_x;
	p->cur_y=twin->cur_y;
	p->start_y=twin->start_y;
	}

void TraceLoad (struct TraceDataStruct *p, struct TwinStruct *twin)
/* Restore Trace output from FarData into hWnd */
	{
	memcpy (twin->screen, p->Display, T_DISPLAY_X*T_DISPLAY_Y);
	twin->cur_x=p->cur_x;
	twin->cur_y=p->cur_y;
	twin->start_y=p->start_y;
	}

void TraceDummy (struct TraceDataStruct *p, char *Text)
/* put Trace output into dummy window and transfer it to TraceList.FarData*/
	{
	if (Text!=NULL)
		{
		TraceLoad (p, dummy_twin);
		simple_write (dummy_twin, Text);
		TraceSave (p, dummy_twin);
		}
	}

void SetProcName (int Tid, long Pid, char *ProcName)
/* change ProcName... */
	{
	struct	TraceDataStruct *p;

	if (!(p=FindTraceData (Tid, Pid)))
		return;
	strcpy (p->ProcName, ProcName);
	trace_list_set (p);
	debug_win_set (p);
	}

struct TraceDataStruct *debug_get_process (Panel_item item)
/*
 *	retrieve the trace data associated with the debug window that owns
 *	item.
 */
	{
	Frame	frame;
	frame=(Frame) xv_get (item, PANEL_CLIENT_DATA);
	return (struct TraceDataStruct *) xv_get (frame, XV_KEY_DATA, DEBUG_ID);
	}

void debug_proc_tracelist (Panel_item item, Event *event)
/*
 *	button notify function
 *	show trace list command frame
 */
	{
	Frame frame;

	frame=(Frame) xv_get (item, PANEL_CLIENT_DATA);
	/*
	if (trace_dlg.frame != frame)
		{
		if (trace_dlg.frame)
			trace_proc_ok (NULL, NULL);
		trace_window (frame);
		}
	else
		trace_proc_ok (NULL, NULL);
	*/
	if (trace_dlg.frame == 0)
		trace_window (console.frame);
	else
		trace_proc_ok (NULL, NULL);
	}
	
void debug_proc_step (Panel_item item, Event *event)
/*
 *	"Step" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	debug_step (p);
	}

void debug_proc_skip (Panel_item item, Event *event)
/*
 *	"Skip" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	if (p->nStatus==T_S_WAITING)
		{
		p->nStatus=T_S_ACTIVE;
		SendRet (p->RTid, p->RPid, 7);
		trace_list_set (p);
		debug_win_set (p);
		}
	}

void debug_proc_fail (Panel_item item, Event *event)
/*
 *	"Fail" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	if (p->nStatus==T_S_WAITING)
		{
		p->nStatus=T_S_ACTIVE;
		SendRet (p->RTid, p->RPid, 3);
		trace_list_set (p);
		debug_win_set (p);
		}
	}

void debug_proc_auto (Panel_item item, Event *event)
/*
 *	"Auto" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	debug_set_auto (p);
	}

void debug_proc_run (Panel_item item, Event *event)
/*
 *	"run" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	debug_set_run (p);
	}

void debug_proc_database (Panel_item item, Event *event)
/*
 *	"Database" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	if (p->nStatus==T_S_WAITING)
		{
		p->nStatus=T_S_PRINTING;
		SendRet (p->RTid, p->RPid, 5);
		}
	}
	
void debug_proc_messagebase (Panel_item item, Event *event)
/*
 *	"Messagebase" button notify function
 */
	{
	struct TraceDataStruct *p;

	p=debug_get_process (item);
	if (p->nStatus==T_S_WAITING)
		{
		p->nStatus=T_S_PRINTING;
		SendRet (p->RTid, p->RPid, 6);
		}
	}

void debug_proc_ignore (Panel_item item, Event *event)
/*
 *	"ignore" button notify function
 */
	{
	struct TraceDataStruct *p;
	struct DebugWindowStruct *dbw;

	p=debug_get_process (item);
	if (p->IgnoreTrace)
		p->IgnoreTrace = FALSE;
	else
		p->IgnoreTrace = TRUE;
	trace_list_set (p);
	}

void debug_proc_close (Panel_item item, Event *event)
/*
 *	"Close" button notify function
 */
	{
	debug_close ((struct DebugWindowStruct *)xv_get (item, PANEL_CLIENT_DATA));
	}

void debug_done_proc (Frame frame)
	{
	struct TraceDataStruct *p;

	p = (struct TraceDataStruct *)xv_get(frame, XV_KEY_DATA, DEBUG_ID);
	debug_close (p->DebugWindow);
	}

void debug_win_set (struct TraceDataStruct *p)
/*
 *	update debug window status
 */
	{
	struct DebugWindowStruct *dbw;
	char	s[200];
	char *c, *q;
	int	n;
	int	fd;

	dbw=p->DebugWindow;
	if (!dbw)
		return;
	sprintf (s, "%6d/%-6d %s : %s", p->Tid, p->Pid, p->ProcName, StateName[p->nStatus]);
	xv_set (dbw->frame, FRAME_LABEL, s, NULL);
	sprintf (s, "%s : %d",p->FileName, p->lFilePos);
	xv_set (dbw->message, PANEL_LABEL_STRING, s, NULL);

	fd = open (p->FileName, O_RDONLY, 0);
	if (fd != -1)
		{
		lseek (fd, p->lFilePos, SEEK_SET);
		read (fd, s, 199);
		s[199]='\0';
#ifdef HAS_STRSEP
		c = s+1;
		q = strsep (&c, " ,()=\\/\"'`\t\n\r;");
		n = c-q;
		if (n<=0) n=1;
#else
		c = strtok (s, " ,()=\\/\"'`\t\n\r;");
		if (c == NULL)
			n = 1;
		else
			n = strlen (c);
		if (n==0) n=1;
#endif
		close (fd);
		}
	xv_set (dbw->textsw, TEXTSW_FILE, p->FileName,  
		TEXTSW_READ_ONLY,	TRUE,
		NULL);
	textsw_possibly_normalize (dbw->textsw, p->lFilePos);
	textsw_set_selection (dbw->textsw, p->lFilePos, p->lFilePos+n, 1);
	}

void debug_win_print (struct TraceDataStruct *p, char *text)
/*
 *	print a line into the debug window or the dummy window
 */
	{
	struct DebugWindowStruct *dbw;
	if (dbw=p->DebugWindow)
		simple_write (dbw->twin, text);
	else
		TraceDummy (p, text);
	}

void debug_win_reset (struct DebugWindowStruct *dbw)
/*
 *	clear debug window...
 */
	{
	simple_clear (dbw->twin);
	xv_set (dbw->frame, FRAME_LABEL, "Debug", NULL);
	xv_set (dbw->message, PANEL_LABEL_STRING, "...", NULL);
	}

void debug_show (struct TraceDataStruct *p)
/*
 *	activate debug window if present,
 *	remap it if iconised
 */
	{
	struct DebugWindowStruct *dbw;

	if (dbw=p->DebugWindow)
		if (xv_get (dbw->frame, FRAME_CLOSED))
			xv_set (dbw->frame, FRAME_CLOSED, FALSE, NULL);
	xv_set (dbw->frame, XV_SHOW, TRUE, NULL);
	}

void debug_open (struct TraceDataStruct *p)
/*
 *	open a debugger window containing the process p
 */
	{
	struct DebugWindowStruct *dbw;
	Panel	panel;
	int	h, y;
	Xv_window	view;

	dbw=(struct DebugWindowStruct *)malloc (sizeof (struct DebugWindowStruct));
	dbw->frame= (Frame) xv_create (console.frame, FRAME,
		FRAME_LABEL,		"Debug",
		FRAME_DONE_PROC,	debug_done_proc,
		XV_HEIGHT,		300,
		XV_KEY_DATA,		DEBUG_ID,	p,
		NULL);

 	dbw->twin=(struct TwinStruct *)malloc (sizeof (struct TwinStruct));
	make_twin_sub (dbw->frame, T_DISPLAY_Y, T_DISPLAY_X, dbw->twin);

	h= (int) xv_get (dbw->frame, XV_HEIGHT);
	xv_set ((dbw->twin)->canvas, XV_HEIGHT, h/2, NULL);

	xv_set (dbw->frame, XV_WIDTH, xv_get ((dbw->twin)->canvas, XV_WIDTH), NULL);

	panel=(Panel) xv_create (dbw->frame, PANEL, 
		WIN_BELOW, 	(dbw->twin)->canvas, 
		XV_X, 		0,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LAYOUT,		PANEL_HORIZONTAL,
		PANEL_LABEL_STRING,	"Close",
		PANEL_NOTIFY_PROC,	debug_proc_close,
		PANEL_CLIENT_DATA,	dbw,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Trace List",
		PANEL_NOTIFY_PROC,	debug_proc_tracelist,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Database",
		PANEL_NOTIFY_PROC,	debug_proc_database,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Messagebase",
		PANEL_NOTIFY_PROC,	debug_proc_messagebase,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	dbw->message= (Panel_item) xv_create (panel, PANEL_MESSAGE,
		PANEL_LABEL_STRING,	"...",
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_NEXT_ROW,		-1,
		PANEL_LABEL_STRING,	"Step",
		PANEL_NOTIFY_PROC,	debug_proc_step,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Skip",
		PANEL_NOTIFY_PROC,	debug_proc_skip,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Fail",
		PANEL_NOTIFY_PROC,	debug_proc_fail,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Auto",
		PANEL_NOTIFY_PROC,	debug_proc_auto,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Run",
		PANEL_NOTIFY_PROC,	debug_proc_run,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	dbw->ignore= (Panel_item)xv_create (panel, PANEL_CHECK_BOX,
		PANEL_LAYOUT,		PANEL_HORIZONTAL,
		PANEL_CHOICE_STRINGS,	"local Ignore Trace", NULL,
		PANEL_NOTIFY_PROC,	debug_proc_ignore,
		PANEL_CLIENT_DATA,	dbw->frame,
		NULL);
	if (p->IgnoreTrace)
		xv_set (dbw->ignore, PANEL_VALUE, 1, NULL);
	else
		xv_set (dbw->ignore, PANEL_VALUE, 0, NULL);

	window_fit_height (panel);
/*	dbw->outwin= make_swin (0,0,T_DISPLAY_Y, T_DISPLAY_X, "Debug output");
 *	dbw->twin=&((struct SwinStruct *)plist_get (object_plist, dbw->outwin))->twin;
 */
/*	xv_set ((dbw->twin)->canvas, WIN_BELOW,	panel, NULL); */
	y= (int) xv_get ((dbw->twin)->canvas, XV_Y);
	h= (int) xv_get (dbw->frame, XV_HEIGHT);
/*	xv_set ((dbw->twin)->canvas, XV_HEIGHT, (h-y)/2, NULL);*/
	dbw->textsw= (Textsw) xv_create (dbw->frame, TEXTSW,
/*		WIN_BELOW,		(dbw->twin)->canvas, */
		WIN_BELOW,		panel,
		XV_X,			0,
		XV_HEIGHT,		(h-y)/2, 
		TEXTSW_BROWSING,	TRUE,
		NULL);
	window_fit (dbw->frame);

	p->DebugWindow=dbw;
	xv_set (dbw->frame, XV_SHOW, TRUE, NULL);
	debug_win_set (p);
	trace_list_set (p);
	}

void debug_close (struct DebugWindowStruct *dbw)
/*
 *	close the debug window dbw.
 */
	{
	struct TraceDataStruct *p;
	if (dbw==NULL)
		return;
	p=(struct TraceDataStruct *)xv_get (dbw->frame, XV_KEY_DATA, DEBUG_ID);
	p->DebugWindow=NULL;
/*	destroy_swin (dbw->outwin);*/
	destroy_twin_sub (dbw->twin);
	free (dbw->twin);
	xv_destroy_safe (dbw->frame);
	free (dbw);
	trace_list_set (p);
	}

void TraceClear ()
/*
 *	clear list of trace processes and remove all open debug windows
 */
	{
	struct TraceDataStruct *p;
	struct list *elem;

	elem=get_first_list_elem (trace_list);
	while (elem)
		{
		p=(struct TraceDataStruct *)elem->data;
		debug_close (p->DebugWindow);
		remove_list_elem (trace_list, elem);
		free (p);
		free (elem);
		elem=get_first_list_elem (elem);
		}
	if (trace_dlg.frame)
		trace_proc_ok (NULL, NULL);
	return;
	}
