/*
 *		DOLINK.C	
 *		Link interfacing routines...
 */
#define noDEBUG

#include "xwio.h"
#include <xview/notify.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#ifdef Linux
#include <linux/b004.h>
#endif Linux
#include "b004.h"
#ifdef USE_LINKINFO
#include <link.h>
#endif
#if (defined USE_LOCAL_PSERVER || defined USE_REMOTE_PSERVER)
#include "pserver.h"
#endif PSERVER

#define CLIENT_ID	17
#define CLIENT_ID2	18
#define BUILDUP_TIME	5

#define noDEBUG_C

extern int ServerTid;
extern int ServerPid;

extern struct SpyStruct *spy;
extern struct SetupDataStruct setupdata;
extern struct ConsoleStruct console;

int out_fd;
int in_fd;
int out_client;
int in_client;

int nMsgBlock=10;

struct BlockQueueStruct 
	{
	struct list *root;
	int count;
	};

struct BlockDataStruct 
	{
	int	type;
	char	Block[1];
	};

struct FileDataStruct 
	{
	int	type;
	int	fd;
	int	tid;
	int	pid;
	int	len;
	int	i;
	};

void file_block (struct LinkStruct *link, struct FileDataStruct *pFile);

#define Q_TYPE_BLOCK	17
#define Q_TYPE_FILE	18

int	flag_sig;

int readit (int fd, char *buf, int len)
	{
	int	i;
	int	l;

	l=len;
	while (len)
		{
		i=read (fd, buf, len);
		if (i==-1)
			{
			if (errno==EAGAIN)
				i=0;
			else
				return -1;
			}
		len-=i;
		buf+=i;
		}
	return l;
	}

int writeit (int fd, char *buf, int len)
	{
	int	i;
	int	l;

	l=len;
	while (len)
		{
		i=write (fd, buf, len);
		if (i==-1)
			{
			if (errno==EAGAIN)
				i=0;
			else
				{
				fprintf (stderr, "writeit detected error %d.\n", errno);
				return -1;
				}
			}
		len-=i;
		buf+=i;
		}
	return l;
	}

void sig_func (int sig, int code, struct sigcontext *scp, char *addr)
/*
 *	Signal handler for SIGUSR1, indicating Reset sequence.
 */
	{
	flag_sig=1;
	}

int reset_sequence_out (struct LinkStruct *link)
/*
 *	wait for end of reset sequence by reading from pipe.
 *	Reading -1 from pipe causes process to terminate.
 */
	{
	int	x;
	readit (link->pipe[0][0], &x, 4);	/* indicate signal received */
	readit (link->pipe[0][0], &x, 4);	/* wait for reset or quit */
	return x;
	}

void DoLink_out (struct LinkStruct *link)
/*
 *	This is the loop executed by the output child process to move
 *	blocks from the parent's pipe to the link.
 */
	{
	int	header[2];
	int 	*buf;
	short	len;
	int	i;

	int flag_sig_out=0;
/*	signal (SIGUSR1, sig_func_out);*/

	buf=(int *)malloc (BLOCK_SIZE);

	for (;;)
		{
		if (!flag_sig_out)
			readit (link->pipe[0][0], header, 8);
		len=bufint16 (header+6) & 0x0FFF;
		if (!flag_sig_out)
			readit (link->pipe[0][0], buf, len);
		if (!flag_sig_out)
			WriteBlock (link->id, (char *)header, (char *)buf);
		if (flag_sig_out)
			if (-1 == reset_sequence_out (link))
				_exit (0);
			else
				flag_sig_out=0;
		}
	}

int DoLink_out_one (struct LinkStruct *link)
/*
 *	This is the loop executed by the output child process to move
 *	blocks from the parent's pipe to the link.
 */
	{
	static char	header[8];
	static char	buf[BLOCK_SIZE];

	short	len;
	int	i;
	static  int dd=0;

	if (link==NULL)
		dd=0;
	else
		{
#ifdef DEBUG
	fprintf (stderr, "DoLink_out_one reading pipe , dd is %d\n", dd);
#endif
		i=read (link->pipe[0][0], header+dd, 8-dd);
		if (i==-1)
			{
			if (errno!=EAGAIN)
				fprintf (stderr, "read from pipe returned error %d.\n", errno);
			i=0;
			}
		if (i<8-dd)
			{
			dd+=i;
			return -1;
			}
		else
			dd=0;

		/*	readit (link->pipe[0][0], header, 8);*/
		len=bufint16 (header+6) & 0x0FFF;
		readit (link->pipe[0][0], buf, len);
#ifdef DEBUG
	fprintf (stderr, "DoLink_out_one read %d. writing block \n",len);
#endif
		WriteBlock (link->id, (char *)header, (char *)buf);
#ifdef DEBUG
	fprintf (stderr, "wrote block \n");
#endif
/*		if (-1 == reset_sequence_out (link))
			_exit (0);
		else
			flag_sig_out=0;
*/
		}
#ifdef DEBUG
	fprintf (stderr, "done out_one, dd is %d \n", dd);
#endif
	return 0;
	}

void send_block (struct LinkStruct *link)
/*
 *	send one block from the output queue over the output pipe.
 */
	{
	struct list *elem;
	struct BlockDataStruct *pBlock;
	int	i;
	short	j;


	if (link->fPause)
		return;
	elem=get_first_list_elem (link->qOut);
	if (elem)
		{
		pBlock = (struct BlockDataStruct *)elem->data;
		if (pBlock->type == Q_TYPE_FILE)
			{
			remove_list_elem (link->qOut, elem);
			free (elem);
			file_block (link, (struct FileDataStruct *)pBlock);
			elem=get_first_list_elem (link->qOut);
			pBlock= (struct BlockDataStruct *)elem->data;
			}
		if (pBlock->type == Q_TYPE_BLOCK)
			{
			j= bufint16 (pBlock->Block+6);
			writeit (link->pipe[0][1], pBlock->Block, j+8);
			remove_list_elem (link->qOut, elem);
			free (elem);
			free (pBlock);
#ifdef DEBUG
	fprintf (stderr, "main process dispatched block of %d, %d blocks left in list. \n",j, count_list_elems (link->qOut));
#endif
			}
		link->fOutReady=0;
		}
	else	{
#ifdef DEBUG
	fprintf (stderr, "send_block on empty list, setting fOutReady.\n");
#endif
		notify_set_output_func ((int)link, NOTIFY_FUNC_NULL, link->pipe[0][1]);
		link->fOutReady=1;
		}
	}

Notify_value out_proc (Notify_client client, int fd)
/*
 *	This is called by the notifier when the output
 *	pipe becomes ready.
 */
	{
#ifdef DEBUG
	fprintf (stderr, "out_proc called by notifier\n");
#endif
	send_block ((struct LinkStruct *)client);
	return NOTIFY_DONE;
	}

void QBlock (struct LinkStruct *link, struct BlockDataStruct *pBlock)
/*
 *	add a Block or list of blocks to the specified Q.
 */
	{
	struct list *elem;

#ifdef DEBUG_C
	fprintf (stderr, ".");
#endif
	elem=(struct list *)malloc (sizeof (struct list));
	elem->data=pBlock;
	assertz_list (link->qOut, elem);
#ifdef DEBUG
	fprintf (stderr, "queued Block, %d in list, fOutReady is %d.\n", count_list_elems (link->qOut), link->fOutReady);
#endif
	if (link->fOutReady)
		{
		notify_set_output_func ((int)link, out_proc, link->pipe[0][1]);
		send_block (link);
		}
	}

void Write_Message (struct LinkStruct *link, int tid, int pid, char *buf, int len)
/*
 *	cut message to BLOCK_SIZE chunks, queue them into the output
 *	queue, start output process.
 */

	{
	char    Header[16];
	int     nCount;
	int     i;
	struct BlockDataStruct *pBuf;

#ifdef DEBUG
	fprintf (stderr, "Write_Message of %d bytes.\n", len);
#endif
	intbuf (pid, Header);
	int16buf ((short)tid , Header+4);
	intbuf (ServerPid, Header+8);
	int16buf ((short)ServerTid, Header+12);
	int16buf (BLOCK_SIZE, Header+6);
	nCount=len/(BLOCK_SIZE-8);
	i=nCount;
	while (i)
		{
		pBuf=(struct BlockDataStruct *) malloc(BLOCK_SIZE+8+sizeof (int));
		pBuf->type=Q_TYPE_BLOCK;
		int16buf (i, Header+14);
		memcpy (pBuf->Block, Header, 16);
		memcpy (pBuf->Block+16, buf+(nCount-i)*(BLOCK_SIZE-8), BLOCK_SIZE-8);
		QBlock (link, pBuf);
		i--;
		}
	pBuf=(struct BlockDataStruct *) malloc(len-nCount*(BLOCK_SIZE-8)+16+sizeof (int));
	pBuf->type=Q_TYPE_BLOCK;
	int16buf ((short)(8+len-nCount*(BLOCK_SIZE-8)), Header+6);
	int16buf (0, Header+14);
	memcpy (pBuf->Block, Header, 16);
	memcpy (pBuf->Block+16, buf+nCount*(BLOCK_SIZE-8), len-nCount*(BLOCK_SIZE-8));
	QBlock (link, pBuf);
	if (spy)
		WriteMsgSpyOut (Header, buf);
	}

Notify_value in_proc(Notify_client client, int fd)
/*
 *	This is the event handler for the input pipe. Collected and
 *	assembled PROLOG messages arrive here.
 *	For protocol on pipe see 'Write_Incoming_Msg'.
 */
	{
	int len;
	int *buf;
#ifdef DEBUG
	fprintf (stderr, "in_proc called by notifier\n");
#endif
	if (!((struct LinkStruct *)client)->fPause)
		{
		readit (fd, &len, sizeof (int));
#ifdef DEBUG
	fprintf (stderr, "in_proc receiving message of %d bytes\n", len);
#endif
		buf=(int *)malloc(len);
		if (NULL==buf)
			{
			fprintf (stderr, "in_proc couldn't malloc %d bytes for buf.\n",len);
			exit (-1);
			}
		readit (fd, buf, len);
#ifdef DEBUG
	fprintf (stderr, "in_proc calling ExecProlog for %d byte msg\n", len);
#endif
		ExecPrologMessage ((char *)buf, len);
		free (buf);
		}
	return NOTIFY_DONE;
        }

void sig_func_in (int sig, int code, struct sigcontext *scp, char *addr)
/*
 *	Signal handler for SIGUSR1, indicating Reset sequence.
 */
	{
	}

int reset_sequence_in (struct LinkStruct *link)
/*
 *	write a word over the pipe to synchronize with parent after reset.
 */
	{
	int	x=0;

	writeit (link->pipe[1][1], &x, sizeof (int));
	writeit (link->pipe[1][1], &x, sizeof (int));
	}

void Write_Incoming_Msg (struct LinkStruct *link, int *p, int tid, int pid, int count)
/*
 *	This function writes a complete block through the pipe to
 *	the XWIO parent process.
 *	Format is 
 *				len,	int
 *		block[0] : 	Pid,	int
 *		block[1] : 	Tid,	int
 *	overall length of block is len bytes.
 *
 */
	{
	int	x;
	int	len=0;
	int 	*q;
	char	header[8];
	int	a;

	q=p;
	do	{
		len+= bufint ((char *)(q+1))*sizeof (int) -8 ;
		if (bufint ((char *)(q+1))==0)
			fprintf (stderr, "Write_incoming found buffer of len 0?.\n");
		q=(int *)(q[0]);
		} while (q);
	len+=8;

	intbuf (pid, header);
	int16buf (tid, header+4);
	int16buf (len, header+6);

#ifdef DEBUG
	fprintf (stderr, "Write_Incoming writing to pipe.\n");
#endif
	writeit (link->pipe[1][1], &len, sizeof (int));
	writeit (link->pipe[1][1], header, 8);
	do	{
		a=bufint ((char *)(p+1))*sizeof (int) -8;
		writeit (link->pipe[1][1], p+2, a);
		q=(int *)(p[0]);
		free (p);
		p=q;
		} while (p);
	}


void DoLink ( struct LinkStruct *link)
/*
 *	DoLink is run by the child process responsible for
 *	reading from the link.
 *	This routine collects blocks from the input link, assembles
 *	them to complete messages and passes them on to the parent
 *	process.
 *	Link protocol is the standard PROLOG protocol, we assume correct
 *	order of blocks, block count in the first block contains the
 *	total number of blocks belonging to this msg.
 *	memory is allocated for the complete message, 
 *	A SIGUSR1 causes reset or quit sequences.
 */
	{
	char 	*header;
	char	*buf;
	int	*p, *q;
	int	msgTid;
	int	msgPid;
	int	count;
	struct pend_msg {
       		int	msgTid;
	       	int	msgPid;
       		int	*p;
		int	*last;
		int	count;
		} *pend;
	struct list *pending;
	struct list *elem;
	int	x;

	flag_sig=0;

	signal (SIGUSR1, sig_func);

	pending=init_list ();
	header = (char *) malloc (2*sizeof (int));
/*	fcntl (link->pipe[0][0], F_SETFL, FNBIO); */
	fcntl (link->pipe[0][0], F_GETFL, &x);
	fcntl (link->pipe[0][0], F_SETFL, x | O_NONBLOCK);
	fcntl (link->pipe[1][1], F_GETFL, &x);
	fcntl (link->pipe[1][1], F_SETFL, x & ~O_NONBLOCK);
	for (;;)
		{
		if (-1==DoLink_out_one (link) && !TestRead (link->id))
			usleep (50);
		if (TestRead (link->id))
			{
			buf=(char *) malloc (BLOCK_SIZE);
			x=ReadBlock (link->id, header, buf);
#ifdef DEBUG
	fprintf (stderr, "read block of %d bytes\n", x);
#endif
			if (x & 3)
				fprintf (stderr, "blocklength %d not word aligned.\n",x);
			if (x<12 || x>BLOCK_SIZE)
				fprintf (stderr, "ReadBlock returned len of %d.\n",x);
			if (flag_sig)
				{
				fprintf (stderr, "DoLink received sig...\\n");
				while (elem=get_first_list_elem (pending))
					{
					remove_list_elem (pending, elem);
					p=((struct pend_msg *)elem->data)->p;
					do	{
						q=p;
						free (p);
						p=(int *)p[0];
						} while (p);
					free (elem->data);
					free (elem);
					}
				reset_sequence_in (link);
				free (buf);
				continue;
				}
	
			count=bufint16 (buf+6);               /* block counter */
			msgPid=bufint (buf);
			msgTid=bufint16 (buf+4) & 0x0000FFFF;
			intbuf (NULL,buf);
			intbuf ((bufint16 (header+6)&0xFFF) /sizeof (int), buf+4);         /* word length of block */
			if (bufint16(header+6)<12)
				fprintf (stderr, "len=%d from %d/%d, count %d for %d/%d.\n", bufint16 (header+6), msgTid, msgPid, count, bufint (header), bufint16(header+4));
	
					/* search for pending msg with same id */
			elem=get_first_list_elem (pending);
			if (elem)
				pend=(struct pend_msg *)elem->data;
			while (elem && (pend->msgTid!=msgTid || pend->msgPid!=msgPid)) 
				{
				elem=get_next_list_elem (elem);
				if (elem)
					pend=(struct pend_msg *)elem->data;
				}
			if (elem) {			/* found pending msg */
				(pend->last)[0]=(int)buf;	/* add block */
				pend->last=(int *)buf;
				if (count==0) 
					{		/* this was the last block */
					remove_list_elem (pending, elem);
					Write_Incoming_Msg (link, pend->p, pend->msgTid, pend->msgPid, pend->count);
					free (elem);
					free (pend);
					}
				}
			else    {			/* block from new msg */
				if (count==0) 
					{		/* single block msg */
					Write_Incoming_Msg (link, (int *)buf, msgTid, msgPid, count);
					}
				else    {		/* new pending msg */
					pend=(struct pend_msg *)malloc (sizeof (struct pend_msg));
       		 	                elem=(struct list *)malloc (sizeof (struct list));
					elem->data=pend;
					pend->p=(int *)buf;
					pend->last=(int *)buf;
					pend->msgTid=msgTid;
					pend->msgPid=msgPid;
					pend->count=count;
					assertz_list (pending, elem);
					}
				}
			}
		}
	}


Notify_value sigchldcatcher (Notify_client client, int pid, union wait *status, struct rusage *rusage)
	{
	if (WIFEXITED (*status))
		{
/*		fprintf (stderr, "sig catcher: pid %d terminated with status %d.\n",pid, status->w_retcode); */
		return NOTIFY_DONE;
		}
/*	fprintf (stderr, "sig not handled.\n"); */
	return NOTIFY_IGNORED;
	}

void pause_link (struct LinkStruct *link)
/*
 *	toggle link pause mode
 */
	{
	if (link->fPause)
		link->fPause=0;
	else
		link->fPause=1;
	}

struct LinkStruct *InitMyLink(char *link_name)
/*
 *      This sets up the child process working on the physical link
 *      interface.
 */
        {
        int i, t=TOUT_A;
        struct LinkStruct *link;
        char *lname = link_name;

#ifdef USE_LOCAL_PSERVER
	/* use Bootlink from pserver */
	lname = "";
	if (-1 ==  init_pserver ())
		{
		fprintf (stderr, "init_pserver failed.\n");
		return NULL;
		}
#endif USE_LOCAL_PSERVER
#ifdef USE_SOCKET_PSERVER
	{
	char *p;

	if (NULL== (p=strchr(link_name, '@')))
		{
		fprintf (stderr, "Illegal remote link %s.\n",link_name);
		return NULL;
		}
	*p = '\0';
	p++;
	if (-1 == connect_to_pserver (p, atoi(link_name)))
		{
		fprintf (stderr, "failed to connect to socket %d at %s.\n", atoi(link_name), p);
		return NULL;
		}
	lname = "";
	}
#endif USE_SOCKET_PSERVER
#ifdef USE_LINKINFO
	{
	struct ext_link_info *li = NULL;
	int	r;

	if (isdigit (link_name[0]))
		{
		r = InfoLink (&li);
		if (r<0 || li == NULL ) {
			simple_write (console.twin, "error in InfoLink.\n");
			return NULL;
			}
		i = atoi (link_name);
		if (i<0 || i>=r) {
			simple_write (console.twin, "Link number out of range");
			return NULL;
			}
		lname = li[i].link_name;
		}
	else
		lname = link_name;
	}
#endif USE_LINKINFO

	link=(struct LinkStruct *)malloc (sizeof (struct LinkStruct));
	if (lname[strlen(lname)-1]=='\n')
		lname[strlen(lname)-1]=0;
	simple_write (console.twin, "opening link ");
	simple_write (console.twin, link_name);
	simple_write (console.twin, ".\n");
        link->id=OpenLink (lname);
        if (link->id < 0)
                {
                free (link);
                return NULL;
                }
#ifdef Linux
	i=ioctl (link->id, B004READABORT, 1);
	i=ioctl (link->id, B004WRITEABORT, 0);
#endif
	i=ioctl (link->id, B004TIMEOUT, &t);

        strncpy (link->name, link_name, 15);
        link->name[15]=0;
	link->qOut=init_list ();
	link->fOutReady=0;
	link->fPause=0;
	return link;
	}

int StartLinkLoops (struct LinkStruct *link)
	{
	int pid;

        pipe (link->pipe[0]);           /* input pipe */
        pipe (link->pipe[1]);           /* output pipe */
        notify_set_output_func ((int)link, out_proc, link->pipe[0][1]);
        notify_set_input_func ((int)link, in_proc, link->pipe[1][0]);

        out_fd=link->pipe[0][1];
        out_client=(int)link;
        in_fd=link->pipe[1][0];
        in_client=(int)link;

        switch (pid=fork())
                {
                case -1:
                        exit (-1);
                case 0:                 /* child process */
			close (link->pipe[1][0]);
			close (link->pipe[0][1]);
                        DoLink (link);
                        _exit (-1);
                default:                /* parent process */
                        link->in_pid=pid;
#ifdef DEBUG_C
	fprintf (stderr, "child pid is %d.\n", pid);
#endif
			notify_set_wait3_func (CLIENT_ID, sigchldcatcher, pid); 
                }
/******
        switch (pid=fork ())
                {
                case -1:        
                        exit (-1);
                case 0:
			close (link->pipe[0][1]);
			close (link->pipe[1][0]);
			close (link->pipe[1][1]);
                        DoLink_out (link);
                        _exit (-1);
		default:
                        link->out_pid=pid;
			notify_set_wait3_func (CLIENT_ID2, sigchldcatcher, pid); 
                }
*******/
	close (link->pipe[0][0]);
	close (link->pipe[1][1]);
	return 1;
        }

void KillLink (struct LinkStruct *link)
/*
 *	Kill link guard, clean up, close pipes...
 *	restore status from after InitMyLink
 */
	{
	struct timeval wait_time;

	if (link)
		{
/* 		fprintf (stderr, "killing link guard (%d)\n", link->in_pid); */
		kill (link->in_pid, SIGKILL);
		wait_time.tv_sec=2;
		wait_time.tv_usec=0;
		select (0, NULL, NULL, NULL, &wait_time);
/*		fprintf (stderr, "cleanup\n"); */
       		notify_set_input_func ((int)link, NOTIFY_FUNC_NULL, link->pipe[1][0]);
       		notify_set_input_func ((int)link, NOTIFY_FUNC_NULL, link->pipe[0][1]);
		close (link->pipe[0][1]);
		close (link->pipe[1][0]);
		link->fOutReady=0;
		link->fPause=0;
		free_list (link->qOut);
		CloseLink (link->id);
		free (link);
		}
	if (!(console.link=InitMyLink (setupdata.linkdriver)))
		{
		fprintf (stderr, "InitMyLink failed for linkdriver %s.\n",setupdata.linkdriver);
		return;
		}
	}
	

/*
 *		Filestr routines...
 *
 */

void file_str (struct LinkStruct *link, int Tid, int Pid, char *file)
        {
        int     nCount, i;
        int	len;
        char	Header[16];
        char	ret[]={
                        43,0,0,0,
                        19,0,0,0,
                        109,115,103,0,
                        43,0,0,0,
                        19,0,0,0,
                        109,115,103,0,
                        7,0,0,0,
                        7,0,0,0,
                        27,0,0,0,
                        19,0,0,0,
                        114,101,116,0,
                        3,0,0,0};
        struct FileDataStruct *pFileBlock;
        struct BlockDataStruct *pBlock;
	struct stat status;
	int	f;
	char *filename;

        if (!Tid && !Pid)
                return;

	filename=find_file (file);
	if (NULL==filename)
		 {
		 char str[200];
		 SendRet (Tid, Pid, -1);
		 sprintf (str, "File not found: %s.",file);
		 error_box (str);
		 return ;
		 }
												      
	if(-1 == stat (filename, &status))
		{
		char str[200];
		SendRet (Tid, Pid, -1);
		sprintf (str, "Can't stat file %s.",filename);
		error_box (str);
		return ;
		}
	len=status.st_size;

	f=open (filename, O_RDONLY);
	if (!f)
		{
		SendRet (Tid, Pid, -1);
		return;
		}
        pBlock=(struct BlockDataStruct *)malloc (BLOCK_SIZE+8+sizeof (int));

        putint (ServerTid, ret+24);
        putint (ServerPid, ret+28);
        memcpy (pBlock->Block+16, ret, 48);
        intbuf (((len/4+1)<<4) | 3L, pBlock->Block+16+44);

        intbuf (Pid, Header);
        int16buf ((short)Tid, Header+4);
        intbuf (ServerPid, Header+8);
        int16buf ((short)ServerTid, Header+12);

        if (len<BLOCK_SIZE-56)
                {       /* one block only, send immediately */

                read (f, pBlock->Block+16+48, len);
                pBlock->Block[16+48+len]='\0';
                while (++len & 0x03L)
                        pBlock->Block[16+48+len]='\0';

                int16buf (8+48+len, Header+6);
                int16buf (0, Header+14);
                memcpy (pBlock->Block, Header, 16);

		pBlock->type=Q_TYPE_BLOCK;
                QBlock (link, pBlock);
		close (f);
                }
        else
                {
                        /* send first block */
                nCount=(len+48)/(BLOCK_SIZE-8);
                i=nCount;
                int16buf (BLOCK_SIZE, Header+6);
                int16buf (i, Header+14);
                memcpy (pBlock->Block, Header, 16);

                read (f, pBlock->Block+16+48, BLOCK_SIZE-(8+48));

		pBlock->type=Q_TYPE_BLOCK;
                QBlock (link, pBlock);
                i--;

                        /* queue descriptor for next blocks */
                pFileBlock=(struct FileDataStruct *)malloc (sizeof (struct FileDataStruct));

		pFileBlock->type=Q_TYPE_FILE;
                pFileBlock->i=i;
                pFileBlock->len=len;
                pFileBlock->tid=Tid;
                pFileBlock->pid=Pid;
		pFileBlock->fd=f;
             /*   pFileBlock->Pos=tell (f);*/

                QBlock (link, (struct BlockDataStruct *)pFileBlock);
                }
        }

void file_block (struct LinkStruct *link, struct FileDataStruct *pFile)
/* send next blocks of file via queue */
        {
        struct BlockDataStruct *pBlock;
        char    Header[16];
        int	rest,pos;
        int     n;

        intbuf (pFile->pid, Header);
        int16buf ((short)pFile->tid, Header+4);
        intbuf (ServerPid, Header+8);
        int16buf ((short)ServerTid, Header+12);
        int16buf (BLOCK_SIZE, Header+6);

        n=min (100, nMsgBlock);         /* queue max 10 blocks each round */

        while (pFile->i && n--)
                {       /* send block via queue 1 */
                pBlock=(struct BlockDataStruct *)malloc (BLOCK_SIZE+8+sizeof (int));

                read (pFile->fd, pBlock->Block+16, BLOCK_SIZE-8);

                int16buf (pFile->i, Header+14);
                memcpy (pBlock->Block, Header, 16);

		pBlock->type=Q_TYPE_BLOCK;
                QBlock (link, pBlock); 

                pFile->i--;
                }
        if (pFile->i)
                {       /* add next block to queue */
             /*   lpFileBlock->lPos=tell (hFile); */
                QBlock (link, (struct BlockDataStruct *)pFile);
                }
        else
                {       /* last block */
                pBlock=(struct BlockDataStruct *)malloc (BLOCK_SIZE+8+sizeof (int));

		pos=lseek (pFile->fd, 0, SEEK_CUR);
                rest=pFile->len - pos;
                read (pFile->fd, pBlock->Block+16, rest);
                pBlock->Block[rest+16]='\0';
                while (++rest & 0x03L)
                        pBlock->Block[rest+16]='\0';

                int16buf (rest+8, Header+6);
                int16buf (0, Header+14);
                memcpy (pBlock->Block, Header, 16);

		pBlock->type=Q_TYPE_BLOCK;
                QBlock (link, pBlock);

		close (pFile->fd);
		free (pFile);
                }
        }
