/************************************************************************/
/*									*/
/*	    EXPROLOG.C							*/
/*                                                                      */
/*	    Prolog message parsing...					*/
/*                                                                      */
/*                                                                      */
/*                                                                      */
/*                                                                      */
/************************************************************************/

#include "xwio.h"
#include "pmsg.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <time.h>
#include <xview/notice.h>

extern int ServerTid;
extern int ServerPid;
extern char *bootfile;
extern struct SpyStruct *spy;
extern struct ConsoleStruct console;
extern struct SetupDataStruct setupdata;
extern char loadfile[];

/* 	definitions for port from WINDOWS: data types */

#define PSTR	char *

#define int_val(A) ((A) >> TAG_LEN_INT)
#define make_int(A) (((A) << TAG_LEN_INT) | INT_TAG)

void dumpmem (int *buf, int bytes)
	{
	int	i, w;

	w=bytes/4;
	fprintf (stderr, "dunmping %d bytes...\n",bytes);
	for (i=0; i<w; i++)
		fprintf (stderr, "%08X  ", bufint (buf+i));
	fprintf (stderr, "done.\n");
	}

#ifndef swint
typedef union {
	int	i;
	char	c[4];
	} Swint;

int swint (int i)
	{
	Swint	x;
	x.c[0]=((Swint)i).c[3];
	x.c[1]=((Swint)i).c[2];
	x.c[2]=((Swint)i).c[1];
	x.c[3]=((Swint)i).c[0];
	return x.i;
	}

typedef union {
	short	i;
	char	c[2];
	} Swi16;

short swi16 (short i)
	{
	Swi16	x;
	x.c[0]=((Swi16)i).c[1];
	x.c[1]=((Swi16)i).c[0];
	return x.i;
	}
#endif

void intbuf (int l, char *buf)
	{
	*(int *)buf=swint (l);
	}

void putint (int i, char *buf)
	{
	*(int *)buf=swint ((i<<4) | 7L);
	}

int bufint (char *buf)
	{
	return swint (*(int *)buf);
	}

int getint (char *buf)
	{
	register int l;
	l= swint ((*(int *)buf));
	if (l<0)
		{
		l>>=4;
		l|=0xF0000000;
		}
	else
		l>>=4;
	return l;
	}

void int16buf (short i, char *buf)
	{
	*(short *)buf=swi16 (i);
	}

int bufint16 (char *buf)
	{
	return (int) swi16(* (short *)buf);
	}


char *SkipString (char *atom)
/* returns pointer to buffer slot folowing an atom with tag in *atom */
	{
	return atom+4*(getint (atom)+1);
	}


void SendSysMessage (int Tid, int Pid, int words, int *msg)
/* send a system message as binary data. */
/* Spy not supported yet. */
	{
	char	buf[BLOCK_SIZE];
	int	i;

	for (i=0; i<words; i++)
		intbuf (msg[i], buf+i*4);
	Write_Message (console.link, Tid, Pid, buf, words*4);
	}



void SendRetStruct (int Tid, int Pid, char *sArg, int len)
/* send return msg */
	{
	char	ret[]={
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			7,0,0,0,
			7,0,0,0,
			27,0,0,0,
			19,0,0,0,
			114,101,116,0,
			11,0,0,0};
	char 	*pBuf;

	if (Tid||Pid)
		{
		putint (ServerTid, ret+24);
		pBuf=(char *)malloc (len+50+4);
		memcpy (pBuf, ret, 48);
		memcpy (pBuf+44, sArg, len);
		while (len & 0x0003)
			pBuf[44+len++]='\0';

		Write_Message (console.link, Tid, Pid, pBuf, 44+len);
		free (pBuf);
		}
	}

void SendRetStr (int Tid, int Pid, char *szArg)
/* send return msg */
	{
	char	ret[]={
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			7,0,0,0,
			7,0,0,0,
			27,0,0,0,
			19,0,0,0,
			114,101,116,0,
			3,0,0,0};
	PSTR	pBuf;
	int	len;

	if (Tid||Pid)
		{
		len=strlen(szArg);
		putint (ServerTid, ret+24);
		pBuf=(PSTR)malloc (len+50+4);
		memcpy (pBuf, ret, 48);
		*(int *)(pBuf+44)=swint((((len/4)+1)<<4)|3L);
		strcpy (pBuf+48, szArg);
		len++;
		while (len & 0x0003)
			pBuf[48+len++]='\0';
		Write_Message (console.link, Tid, Pid, pBuf, 48+len);
		free (pBuf);
		}
	}

void SendRet (int Tid, int Pid, int arg)
/* send return msg */
	{
	char	ret[]={	
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			7,0,0,0,
			7,0,0,0,
			27,0,0,0,
			19,0,0,0,
			114,101,116,0,
			7,0,0,0};

	if (Tid||Pid)
		{
		putint (ServerTid, ret+24);
		putint (arg, ret+44);
		Write_Message (console.link, Tid, Pid, ret, 48);
		}
	}

void ExecPrologMessage (char *buf, int len)
/* interpret and execute Prolog Message */
	{
	int	i;
	int	a,b,c,d;
	int 	Tid;
	int	Pid;
	char	str[200];
	int	hWin;

	if (spy)
		WriteMsgSpyIn (buf);
	if (bufint(buf+16)==PM_PRINT)
		{
		if ((buf[28]&0x0F)== 7)
			sprintf (str, "%d", getint (buf+28));
		else
			if ((buf[28]&0x0F)== 3)
				strcpy (str, buf+32);
			else
				PrintMsg (str, buf+28);
		hWin=getint (buf+24);

		if (hWin<6)
			simple_write (console.twin, str);
			/* printf (str); */
		else
			simple_write_pid (hWin, str);

		}
	else
		{
		Tid=getint (buf+32);
		Pid=getint (buf+36);
		switch (bufint(buf+48))
			{
			case PM_SERVERT:
				SendRetStr (Tid, Pid, "XWIO");
				break;

			case PM_PRINT:
				a=getint (buf+56);

				if (a<6)
					simple_write (console.twin, buf+64);
				else
					write_object (a, buf+64);
				SendRet (Tid, Pid, 0);
				break;

			case PM_GETNIF:
				SendRetStr (Tid, Pid, setupdata.networkfile);
				break;

			case PM_CURSOR:
				hWin=getint (buf+56);
				a=getint (buf+60);
				b=getint (buf+64);
				if (hWin<6)
					fprintf (stderr, "Cursor %d/%d for window %d.\n",a,b,hWin);
				else
					simple_cursor (hWin, b, a);
				SendRet (Tid, Pid, 0);
				break;

			case PM_MAKEWIN:
				a=getint (buf+56);
				b=getint (buf+60);
				c=getint (buf+64);
				d=getint (buf+68);
				hWin=make_swin (a, b, c, d, buf+76);
				SendRet (Tid, Pid, hWin);
				break;

			case PM_KILLWIN:
				a=getint  (buf+56);
				if (a>5)
					destroy_object (a);
				SendRet (Tid, Pid, 0);
				break;

			case PM_CLEARWIN:
				hWin=getint (buf+56);
				if (hWin<6)
					simple_clear (console.twin);
				else
					clear_object (hWin);
				SendRet (Tid, Pid, hWin);
				break;

			case PM_SWITCHWIN:
				hWin=getint (buf+56);
			/*####*/
				SendRet (Tid, Pid, hWin);
				break;

			case PM_WINTEXT:
				hWin=getint (buf+56);
				if (hWin<6)
					xv_set (console.frame, 
						FRAME_LABEL,	buf+64,
						NULL);
				else
					text_object (hWin, buf+64);
				SendRet (Tid, Pid, hWin);
				break;

			case PM_ERROR:
				error_box (buf+60);
				SendRet (Tid, Pid, 1);
				break;

			case PM_STARTUP:
				RunApp (loadfile);
				break;

			case PM_PROC:
				switch (bufint (buf+52))
					{
					case PM_PROCESS:
						i=AddProcess (getint  (buf+56), getint	(buf+60), buf+68);
						SendRet (Tid, Pid, i);
						break;

					case PM_PROCDEL:
						i=DelProcess (getint  (buf+56), getint	(buf+60));
						SendRet (Tid, Pid, i);
						break;
					}
				break;

			case PM_TRACE:
				switch (bufint (buf+52))
					{
					case TRACEON:
						i=TraceOn (getint(buf+56), getint  (buf+60));
						SendRet (Tid, Pid, i);
						break;

					case TRACEOF:
						i=TraceOff (getint (buf+56), getint  (buf+60));
						SendRet (Tid, Pid, i);
						break;

					case TRACE:
						i=getint  (buf+68);
						i=Trace (Tid, Pid, getint  (buf+56), getint  (buf+60), getint  (buf+64), buf+72, buf+76+i*4);
						if (i!=-1)
							SendRet (Tid, Pid, i);
						break;

					case TRACEPR:
						i=TracePrint (getint  (buf+56), getint	(buf+60), buf+68);
						SendRet (Tid, Pid, i);
						break;
					}
				break;

			case PM_FILE:
				switch (bufint (buf+52))
					{
					case FILESTR:
						strlwr (buf+64);
						file_str (console.link, Tid, Pid, buf+64);
						break;

					case FILEPOS:
						SendRet (Tid, Pid, file_pos (getint  (buf+56), getint  (buf+60)));
						break;

					case FILEDAT:
						SendRet (Tid, Pid, file_date (buf+60));
						break;

					case FILESIZ:
						SendRet (Tid, Pid, file_size (buf+60));
						break;

					case FILEOUT:
						{
						char *p;
						p=SkipString (buf+56);
						SendRet (Tid, Pid, file_out (buf+60, p+4, 4*getint (p)));
						break;
						}
					}
				break;

			case PM_WRITEF:
				a=getint  (buf+56);
				write_n_object (a, getint  (buf+60), buf+68);
				SendRet (Tid, Pid, 0);
				break;

			case PM_READ:
				switch (bufint(buf+52))
					{
					case READLN:
						a=getint  (buf+56);
						if (a<6) 
							simple_readln (console.twin, Tid, Pid);
						else
							readln_object (a, Tid, Pid);
						break;

					case READF:
						a=getint  (buf+56);
						if (a<6)
							SendRet (Tid, Pid, -1);
						read_object (a, getint  (buf+60), Tid, Pid);
						break;

					case READCH:
						a=getint (buf+56);
						if (a<6)
							simple_getch (console.twin, Tid, Pid);
						else
							readch_object (a, Tid, Pid);
						break;
					}
				break;

			case PM_OPEN:
				switch (bufint (buf+52))
					{
					case OREAD:
						a=open_file (buf+60, OM_READ);
						break;
					case OAPPEND:
						a=open_file (buf+60, OM_APPEND);
						break;
					case OWRITE:
						a=open_file (buf+60, OM_WRITE);
						break;
					case OMODIFY:
						a=open_file (buf+60, OM_MODIFY);
						break;
					}
				SendRet (Tid, Pid, a);
				break;

			case PM_CLOSEF:
				a=getint  (buf+56);
				SendRet (Tid, Pid, destroy_object (a));
				break;

			case PM_EOFILE:
				a=getint (buf+56);
				SendRet (Tid, Pid, file_eof (a));
				break;

			case PM_GETROW:
				hWin=getint (buf+56);
				if (hWin<6)
					fprintf (stderr, "Getrow for window %d.\n",hWin);
				else
					SendRet (Tid, Pid, simple_getrow (hWin));
				break;

			case PM_GETC:
				switch (bufint (buf+52))
					{
					case GETCOL:
						hWin=getint (buf+56);
						if (hWin<6)
							fprintf (stderr, "Getcol for window %d.\n",hWin);
						else
							SendRet (Tid, Pid, simple_getcolumn (hWin));
						break;
					case GETCH:
						a=getint  (buf+56);
						readch_object (a, Tid, Pid);
						break;
					case GETCDIR:
					case GETCWD:
						{
						char s[255];
						if (getcwd (s,255))
							SendRetStr (Tid, Pid, s);
						else
							SendRet (Tid, Pid, -1);
						break;
						}
					/*
					case GETCDIR:
						{
						char s[255];
						a=getcurdir (getint (buf+56) -1, s);
						if (a==-1)
							SendRet (Tid, Pid, -1);
						else
							SendRetStr (Tid, Pid, s);
						break;
						}
						*/
					}
				break;

			case PM_HOSTEXEC:
			case PM_WINEXEC:
				SendRet (Tid, Pid, host_exec (buf+60));
				break;

			case PM_SPYON:
				if (!spy)
					InitMsgSpy (setupdata.spyfile);
				SendRet (Tid, Pid, 0);
				break;

			case PM_GETENV:
				{
				char *p;

				p=getenv (buf+60);
				if (p)
					SendRetStr (Tid, Pid, p);
				else
					SendRet (Tid, Pid, -1);
				break;
				}
			case PM_TIME:
				{
				struct tm *t;
				struct timeb tb;
				char str[100];
				ftime (&tb);
				t=localtime(&(tb.time));
				if (t)
					{
					intbuf((int)0x4B, str);
					intbuf((int)0x23, str+4);
					intbuf((int)PM_TIME, str+8);
					intbuf(0L,str+12);
					putint((int)t->tm_hour, str+16);
					putint((int)t->tm_min, str+20);
					putint((int)t->tm_sec, str+24);
					putint((int)(tb.millitm/10), str+28);
					SendRetStruct (Tid, Pid, str, 32);
					}
				else
					SendRet (Tid, Pid, -1);
				break;
				}

			case PM_MKGRAPH:
				a=make_gwin (getint  (buf+56), getint  (buf+60), buf+68);
				SendRet (Tid, Pid, a);
				break;

			case PM_POLYGON:
				a=getint  (buf+56);
				SendRet (Tid, Pid, DrawPolygon (a, buf+60));
				break;
				
			case PM_ELLIPSE:
				a=getint  (buf+56);
				SendRet (Tid, Pid, DrawEllipse (a, buf+60));
				break;

			case PM_ENDDRAW:
				a=getint  (buf+56);
				SendRet (Tid, Pid, EndDraw (a));
				break;

			case PM_EXIT:
				{
				Xv_notice	notice;
				int		notice_stat;

				a=getint  (buf+56);
				sprintf (str, "Terminated by %d, %ld, Returncode %ld.\nRestart Server ?", Tid, Pid, a);

				notice= xv_create (console.frame, NOTICE,
					NOTICE_MESSAGE_STRING,  str,
					NOTICE_BUTTON_YES,      "Yes",
					NOTICE_BUTTON_NO,       "No",
					NOTICE_STATUS,          &notice_stat,
					XV_SHOW,                TRUE,
					NULL);
				xv_destroy_safe (notice);

				if (notice_stat == NOTICE_NO)
					quit ();
				else
					{
					cleanup ();
					KillLink (console.link);
					if (-1==Reset (console.link))
						simple_write (console.twin, "XWIO: Reset failed.\n");
					else
						StartLinkLoops (console.link);
					}
				break;
				}
			case PM_SOUND:
				XBell (console.dpy, 0);
				SendRet (Tid, Pid, -1);
				break;

			case PM_CHDIR:
				{
				char s[255];
				strcpy (s, buf+60);
				SendRet (Tid, Pid, (int)chdir (s));
				break;
				}
			case PM_MKDIR:
				{
				char s[255];
				strcpy (s, buf+60);
				SendRet (Tid, Pid, (int)mkdir (s, 0));
				break;
				}
			case PM_RMDIR:
				{
				char s[255];
				strcpy (s, buf+60);
				SendRet (Tid, Pid, (int)rmdir (s));
				break;
				}

			case PM_DELETEF:
				SendRet (Tid, Pid, unlink (buf+60));
				break;

			case PM_MSGBOX:
				{
				Frame win;
				a=getint (buf+56);
				if (a<6)
					win = console.frame;
				else
					win = object_getwin (a);
				make_msgbox (win, getint (buf+60), buf+68, SkipString (buf+64) +4, Tid, Pid);
				break;
				}
#ifdef SOWEIT__
			case PM_GETDISK:
				SendRet (Tid, Pid, (int)_getdrive());
				break;

			case PM_SETDISK:
				a=(int)_chdrive (getint  (buf+56));
				if (a)
					SendRet (Tid, Pid, -1);
				else
					SendRet (Tid, Pid, 0);
				break;

			case PM_GPROFILE:
				{
				LPSTR app, key;

				app=SkipString (buf+56);
				key=SkipString (app);

				switch (bufint32 (buf+52))
					{

					case PROFINT:
						if (*(buf+60))
							a=GetPrivateProfileInt (app+4, key+4, getint  (SkipString (key)), buf+60);
						else
							a=GetProfileInt (app+4, key+4, getint  (SkipString (key)));
						SendRet (Tid, Pid, a);
						break;
					case PROFSTR:
						{
						char b[255];
						b[0]=0;
						if (*(buf+60))
							GetPrivateProfileString (app+4, key+4, SkipString (key)+4, (LPSTR)b, 255, buf+60);
						else
							GetProfileString (app+4, key+4, SkipString (key)+4, (LPSTR)b, 255);
						SendRetStr (Tid, Pid, b);
						break;
						}
					}
				break;
				}

			case PM_SPROFILE:
				{
				LPSTR app, key;
				char b[30];
				char far *s;

				app=SkipString (buf+56);
				key=SkipString (app);

				switch (bufint32 (buf+52))
					{
					case PROFINT:
						sprintf (b, "%ld", getint  (SkipString (key)));
						s=(LPSTR)b;
						break;
					case PROFSTR:
						s=SkipString (key)+4;
						break;
					}
				if (*(buf+60))
					a=(int)WritePrivateProfileString (app+4, key+4, s, buf+60);
				else
					a=(int)WriteProfileString (app+4, key+4, s);
				SendRet (Tid, Pid, a);
				break;
				}

			case PM_DDEINIT:
				DDEInit (hWnd, Tid, Pid, buf+60, SkipString (buf+56)+4);
				break;

			case PM_DDEEXEC:
				hWin=getint  (buf+56);
				DDEExecute (hWin, Tid, Pid, buf+64);
				break;

			case PM_DDETERM:
				hWin=getint  (buf+56);
				DDETerminate (hWin, Tid, Pid);
				break;

			case PM_DDEDATA:
				hWin=getint  (buf+56);
				DDEData (hWin, Tid, Pid, buf+64);
				break;

#endif

			case PM_RET:
				break;
			default:
				fprintf(stderr, "unknown PMsg %s from T/P %X/%X total %u Bytes. \n", buf+48,Tid,Pid,len);
				dumpmem (buf, len);
		/*		Error ("Unidentified PROLOG message");*/
			}
		}
	}


void RunApp (char *p)
/* launch application on transputer... */
	{
	char	run[]={
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			43,0,0,0,
			19,0,0,0,
			109,115,103,0,
			7,0,0,0,
			7,0,0,0,
			27,0,0,0,
			19,0,0,0,
			114,117,110,0,
			3,0,0,0};
	char *pBuf;

	if (!p[0])
		return ;
	if (p[strlen(p)-1]=='\n')
		p[strlen(p)-1]=0;
	if (p[0] != '/')
		if (strrchr (p, '.'))
			if (!strcmp (".lkd",strrchr (p,'.')))
				*(strrchr (p, '.'))='\0';
	putint (ServerTid, run+24);
	pBuf=malloc (strlen (p)+50);
	if (pBuf == NULL )
		return;
	memcpy (pBuf, run, 48);
	intbuf((((strlen(p)/4)+1)<<4)|3L, pBuf+44);
	strcpy (pBuf+48, p);
	Write_Message (console.link, 1, 1L, pBuf, 48+4*((int)(strlen(p)/4)+1));
	free (pBuf);
	}
