#include <X11/Xlib.h>
#include <xview/xview.h>
#include <xview/canvas.h>
#include <xview/xv_xrect.h>
#include <xview/scrollbar.h>
#include <stdlib.h>
#include <memory.h>

#include "xwio.h"

extern struct pidref *object_plist;
extern	GC	gc;
extern struct ConsoleStruct console;
extern Menu twin_menu;


#define	GTYPE_FPOLY	1
#define	GTYPE_POLY	2
#define	GTYPE_ELLIPSE	3

struct DrawDataStruct {
	int	type;
	char 	data[1];
	};
struct DrawPolyStruct	{
	int	type;
	int	bordercol;
	int	fillcol;
	int	count;
	int	p[1];
	};
struct	DrawEllipseStruct {
	int	type;
	int	bordercol;
	int	fillcol;
	int	x1;
	int	y1;
	int 	dx;
	int	dy;
	};
struct GwinStruct {
	Canvas	canvas;
	int	size_x;
	int	size_y;
	struct list *display_list;
	struct list *draw_list;
	int 	flags;
	};

int	*colors=NULL;
int	max_colors=10;
int	num_colors=8;
char	**col_name;
char	*col_default[]={
		"black",
		"dark gray",
		"gray",
		"light gray",
		"white",
		"red",
		"green",
		"blue",
		""};
int	col_num_default[]={ 4,3,2,1,0,6,7,8};
int	*col_num;
int	col_opaque=5;

void load_color_setup ()
	{
	FILE	*f;
	char	str[300];
	char	s[300];
	char	*key, *cont, *name;
	int	i=0;

	strcpy (str, getenv ("HOME"));
	strcat (str, "/.xwiorc");
	
	f=fopen (str, "r");
	if (f)	{
		while (!feof(f))
			{
			fgets (s, 300, f);
			if (getkey (s, &key, &cont))
				{
				if (!strcmp (key, "color"))
					i++;
				}
			}
		fclose (f);
		if (i)
			{
			num_colors=i;
			col_num=(int *)malloc (num_colors * sizeof (int));
			memset ((char *)col_num, 0, num_colors * sizeof (int));
			col_name=(char **)malloc (num_colors*sizeof (char *));
			memset ((char *)col_name, 0, num_colors * sizeof (char *));
			i=0;
			f=fopen (str, "r");
			if (f)	{
				while (!feof(f))
					{
					fgets (s, 300, f);
					if (getkey (s, &key, &cont))
						if (!strcmp (key, "color"))
							{
							if (name=strchr (cont, ':'))
								{
								*name='\0';
								name++;
								if (!strstr(name, "NULL"))
									{
									col_num[i]=atoi(cont);
									col_name[i]=(char *)malloc (strlen (name)+1);
									if (strrchr (name, '\n'))
										*(strrchr (name, '\n'))='\0';
									strcpy (col_name[i], name);
									i++;
									}
								else
									col_opaque=atoi (cont);
								}
							}
					}
				}
			fclose (f);
			return;
			}
		}
	col_num=col_num_default;
	col_name=col_default;
        }


void gwin_init_colors ()
/*
 *	Parse colors...
 */
	{
	int	i;
	XColor	coldef;
	Colormap	default_cmap;

	load_color_setup ();
	default_cmap=DefaultColormap (console.dpy, console.screen_no);
	colors= (int *)malloc (max_colors*sizeof (int));
	for (i=0; i<num_colors; i++)	
		{
		if (col_name[i]!=NULL)
			{
			if (!XParseColor (console.dpy, default_cmap, col_name[i], &coldef))
				{
				fprintf (stderr, "color %s not in database.\n", col_name[i]);
				continue;
				}
			if (!XAllocColor (console.dpy, default_cmap, &coldef))
				{
				fprintf (stderr, "cannot allocate color %s.\n", col_name[i]);
				continue;
				}
			colors[col_num[i]]=coldef.pixel;
			}
		else
			colors[col_num[i]]=0;
		}
	}


void gwin_event_proc (Xv_window window, Event *event)
	{
	struct GwinStruct *gwin;

	switch (event_action (event))
		{
		case ACTION_MENU:
			menu_show (twin_menu, window, event, NULL);
			break;
		}
	return;
	}

void display_filled_poly (Display *dpy, Window xwin, GC gc, struct DrawPolyStruct *poly, int f)
	{
	int 	i;
	XPoint 	*points;
	points=(XPoint *)malloc (poly->count*sizeof (XPoint));
	for (i=0; i<poly->count; i++)
		{
		points[i].x=f*poly->p[2*i];
		points[i].y=f*poly->p[2*i+1];
		}
	if (poly->fillcol != col_opaque)
		{
		XSetForeground (dpy, gc, colors[poly->fillcol]);
		XFillPolygon (dpy, xwin, gc, points, poly->count, Complex, CoordModeOrigin);
		}
	if (poly->bordercol != col_opaque)
		{
		XSetForeground (dpy, gc, colors[poly->bordercol]);
		XDrawLines (dpy, xwin, gc, points, poly->count, CoordModeOrigin);
		}
	free (points);
	}

void display_ellipse (Display *dpy, Window xwin, GC gc, struct DrawEllipseStruct *elli, int f)
	{
	if (elli->fillcol != col_opaque)
		{
		XSetForeground (dpy, gc, colors[elli->fillcol]);
		XFillArc (dpy, xwin, gc, f*elli->x1, f*elli->y1, f*elli->dx, f*elli->dy, 0, 64*360);
		}
	if (elli->bordercol != col_opaque)
		{
		XSetForeground (dpy, gc, colors[elli->bordercol]);
		XDrawArc (dpy, xwin, gc, f*elli->x1, f*elli->y1, f*elli->dx, f*elli->dy, 0, 64*360);
		}
	}

void gwin_display_elem (Display *dpy, Window xwin, GC gc, struct DrawDataStruct *g, int f)
/*
 *	Draw element of display_list into window...
 */
	{
	switch (g->type)
 		{
		case GTYPE_ELLIPSE:
			display_ellipse (dpy, xwin, gc, (struct DrawEllipseStruct *)g, f);
			break;
		case GTYPE_FPOLY:
			display_filled_poly (dpy, xwin, gc, (struct DrawPolyStruct *)g, f);
			break;
		}
	}

void gwin_repaint_proc (Canvas canvas, Xv_window paint_window, Display *dpy, Window xwin, Xv_xrectlist *xrects)
	{
	struct GwinStruct *gwin;
	struct list *elem;
	int	f;

/*	GC gc=DefaultGC (console.dpy, console.screen_no);*/

	XClearWindow (dpy, xwin);
	gwin= (struct GwinStruct *)xv_get (paint_window, XV_KEY_DATA, GWIN_ID);
	f=min ((xv_get (paint_window, XV_WIDTH) / gwin->size_x),(xv_get (paint_window, XV_HEIGHT) / gwin->size_y));
	elem=get_first_list_elem (gwin->display_list);
	while (elem)
		{
		gwin_display_elem (dpy, xwin, gc, (struct DrawDataStruct *)(elem->data), f);
		elem=get_next_list_elem (elem);
		}
	}

void graphics_refresh (struct GwinStruct *gwin)
	{
	gwin_repaint_proc (gwin->canvas, canvas_paint_window (gwin->canvas), (Display *)xv_get (gwin->canvas,XV_DISPLAY), xv_get (canvas_paint_window (gwin->canvas), XV_XID), (Xv_xrectlist *)NULL);
        }

void gwin_resize_proc (Canvas canvas, int width, int height)
	{
	struct GwinStruct *gwin;
	gwin= (struct GwinStruct *)xv_get (canvas, XV_KEY_DATA, GWIN_ID);
	graphics_refresh (gwin);
	}
	
struct GwinStruct *make_graphwin_sub (Frame frame, int x, int y)
/*
 *	create a graphics window to fit into the frame...
 */
	{
	struct GwinStruct *gwin;
	if (colors==NULL)
		gwin_init_colors ();
	gwin=(struct GwinStruct *)malloc (sizeof (struct GwinStruct));
	gwin->canvas= (Canvas) xv_create (frame, CANVAS,
		CANVAS_X_PAINT_WINDOW,	TRUE,
		CANVAS_RESIZE_PROC,	gwin_resize_proc,
		CANVAS_REPAINT_PROC,	gwin_repaint_proc,
		CANVAS_AUTO_SHRINK,	TRUE,
		CANVAS_AUTO_EXPAND,	TRUE,
		XV_KEY_DATA,	GWIN_ID,	gwin,
		NULL);
		
	xv_set (canvas_paint_window (gwin->canvas), 
		WIN_EVENT_PROC,		gwin_event_proc,
		WIN_CONSUME_EVENTS,
			WIN_MOUSE_BUTTONS,
			NULL,
		XV_KEY_DATA,	GWIN_ID,	gwin,
		NULL);

	gwin->size_x=x;
	gwin->size_y=y;
	gwin->draw_list=init_list ();
	gwin->display_list=init_list ();
	return gwin;
	}

void gwin_done_proc (Frame frame)
	{}

int make_gwin (int x, int y, char *title)
/*
 *	create frame for graphics window, install subwins
 *	return pid of new object.
 */
	{
	struct GraphwinStruct *g;
	g=(struct GraphwinStruct *)malloc (sizeof (struct GraphwinStruct));
	g->frame= (Frame) xv_create (console.frame, FRAME,
		FRAME_LABEL,	title,
		FRAME_DONE_PROC,	gwin_done_proc,
		NULL);
	g->gwin=make_graphwin_sub (g->frame, x, y);
	g->magic=PID_MAGIC_GWIN;

	window_fit (g->frame);
	xv_set (g->frame, XV_SHOW, TRUE, NULL);

	return plist_add (object_plist, g);
	}

void destroy_graphwin_sub (struct GwinStruct *gwin)
/*
 *	Kill graphics window...
 */
	{
	free_list (gwin->draw_list);
	free_list (gwin->display_list);
	xv_destroy_safe (gwin->canvas);
	free (gwin);
	}

int destroy_gwin (struct GraphwinStruct *g)
/*
 *	remove graphics window frame...
 */
	{
	destroy_graphwin_sub (g->gwin);
	xv_destroy_safe (g->frame);
	free (g);
	return TRUE;
	}

struct GwinStruct *get_gwin (int pid)
/*
 *	get gwin subwindow handle from pid.
 */
	{
	struct GraphwinStruct *g;

	g=(struct GraphwinStruct *)plist_get (object_plist, pid);
	if (!g )
		{
		printf ("gwin: pid %d not found.\n",pid);
		return NULL;
		}
	else
		if (g->magic != PID_MAGIC_GWIN)
			{
			printf ("gwin: illegal magic for pid %d.\n",pid);
			return NULL;
			}
		else
			return g->gwin;
	}

void add_draw_elem (struct GwinStruct *gwin, void *d)
/*
 *	add an item to the drawlist
 */
	{
	struct list *elem;

	elem=(struct list *)malloc (sizeof (struct list));
	elem->data=d;
	assertz_list (gwin->draw_list, elem);
	}

int DrawPolygon (int oid, char *buf)
/* draw polygon from PROLOG message */
	{
	int	n,i;
	struct GwinStruct *gwin;
	struct DrawPolyStruct *poly;

	if (!(gwin=get_gwin (oid)))
		return -1;
	n=(bufint (buf+4)>>2)/2;
	poly=(struct DrawPolyStruct *) malloc (sizeof (struct DrawPolyStruct)+2*n*sizeof (int));
	if (NULL==poly)
		return -1;
	poly->type=GTYPE_FPOLY;
	poly->bordercol=0;
	poly->fillcol=getint (buf);
	poly->count=n;
	for (i=0; i<n; i++)
		{
		poly->p[2*i]=getint (buf+8+i*8);
		poly->p[2*i+1]=getint (buf+12+i*8);
		}
	add_draw_elem (gwin, poly);
	return n;
	}

int DrawEllipse (int oid, char *buf)
/* draw ellipse from PROLOG message */
	{
	struct GwinStruct *gwin;
	struct	DrawEllipseStruct *elli;

	if (!(gwin=get_gwin (oid)))
		return -1;
	elli=(struct DrawEllipseStruct *) malloc (sizeof (struct DrawEllipseStruct));
	if (NULL==elli)
		return -1;
	elli->type=GTYPE_ELLIPSE;
	elli->fillcol=getint (buf);
	elli->bordercol=0;
	elli->x1=getint (buf+4);
	elli->y1=getint (buf+8);
	elli->dx=getint (buf+12) -elli->x1;
	elli->dy=getint (buf+16) -elli->y1;
	add_draw_elem (gwin, elli);
	return 0;
        }

int EndDraw (int oid)
/*
 *	close metafile,
 *	free displaylist, use drawlist as new displaylist
 */
	{
	struct GwinStruct *gwin;

	if (!(gwin=get_gwin (oid)))
		return -1;

	free_list (gwin->display_list);
	gwin->display_list=gwin->draw_list;
	gwin->draw_list=init_list ();
	graphics_refresh (gwin);
	return 0;
	}
