/************************************************************************/
/*									*/
/*				LIST.C					*/
/*									*/
/************************************************************************/

/**************	Includes ***********************/

#include "xwio.h"

/*
 *      functions for double link list.
 *      This replaces parts of list.c
 */


struct list *init_list ()
/*
 *      initialize list structure
 *      return pointer to list root.
 */
        {
        struct list *root;

        root=(struct list *)malloc (sizeof (struct list));
        root->data=NULL;
        root->next=NULL;
        root->prev=NULL;
        return root;
        }

int is_empty_list (struct list *root)
/*
 *      check, if list is empty
 */
        {
        if (root->next)
                return FALSE;
        else
                return TRUE;
        }

struct list *get_first_list_elem (struct list *root)
/*
 *      return pointer to first elem in list (not to its data!)
 */
        {
	if (root == NULL)
		return NULL;
        return root->next;
        }

struct list *get_next_list_elem (struct list *elem)
/*
 *      return pointer to last elem in list (not to its data)
 */
        {
	if (elem == NULL)
		return NULL;
        return elem->next;
        }

struct list *get_prev_list_elem (struct list *elem)
/*
 *      get previous list element
 */
        {
	if (elem == NULL)
		return NULL;
        return elem->prev;
        }

void asserta_list (struct list *root, struct list *elem)
/*
 *      insert list elem as first elem of list
 */
        {
        elem->next=root->next;
        elem->prev=NULL;
        if (elem->next ==NULL)
                root->prev=elem;
        else
                (elem->next)->prev=elem;
        root->next=elem;
        }

void assertz_list (struct list *root, struct list *elem)
/*
 *      append elem to list
 */
        {
        elem->prev=root->prev;
        elem->next=NULL;
        if (elem->prev ==NULL)
                root->next=elem;
        else
                (elem->prev)->next=elem;
        root->prev=elem;
        }

void remove_list_elem (struct list *root, struct list *elem)
/*
 *      remove list elem, adjust pointers
 *      this routine does not free the list elem memory!
 */
        {
        if (elem->next==NULL)
                root->prev=elem->prev;
        else
                (elem->next)->prev=elem->prev;
        if (elem->prev==NULL)
                root->next=elem->next;
        else
                (elem->prev)->next=elem->next;
        }

void free_list (struct list *root)
/*
 *	clear list, free all list elements,
 *	free all data items, assuming elem->data is a 
 *	pointer to memory block allocated with malloc().
 *	The root elem is also freed.
 */
	{
	struct list *p, *q;
	if (root==NULL)
		{
		fprintf (stderr, "free_list called for root=NULL");
		return;
		}
	p=root->next;
	while (p)
		{
		free (p->data);
		q=p->next;
		free (p);
		p=q;
		}
	free(root);
	}

struct list *find_list_id (struct list *root, int id)
/*
 *      find list elem with first word of data block (elem->data)
 *      identical to id.
 */
        {
        struct list *p;

        p=root->next;
        while (p)
                {
                if (*(int *)(p->data)==id)
                        return p;
                p=p->next;
                }
        return NULL;
        }

struct list *find_list_ptr (struct list *root, void *ptr)
/*
 *      find list elem with data pointer ptr
 */
        {
        struct list *p;

        p=root->next;
        while (p)
                {
                if (p->data==ptr)
                        return p;
                p=p->next;
                }
        return NULL;
        }

int list_member (struct list *root, struct list *elem)
/*
 *      check for elem in list.
 */
        {
        struct list *p;

        p=root->next;
        while (p)
                {
                if (p==elem)
                        return TRUE;
                p=p->next;
                }
        return FALSE;
        }

int count_list_elems (struct list *root)
/*
 *      return number of elements in list
 */
        {
        struct list *p;
        int	i=0;

        p=root->next;
        while (p)
                {
                i++;
                p=p->next;
                }
        return i;
        }

