#include "xwio.h"
#include <xview/xview.h>

extern struct pidref *object_plist;

#define MB_OK		0
#define MB_OKCANCEL	1
#define MB_ABREIG	2
#define MB_YESNOCANCEL	3
#define MB_YESNO	4
#define MB_RETRYCANCEL	5

#define MB_ID_OK	1
#define MB_ID_CANCEL	2
#define MB_ID_ABORT	3
#define MB_ID_RETRY	4
#define MB_ID_IGNORE	5
#define MB_ID_YES	6
#define MB_ID_NO	7

void notice_event_proc (Xv_Notice notice, int value, Event *event)
	{
	struct MsgboxStruct *m;

	m = (struct MsgboxStruct *) xv_get (notice, XV_KEY_DATA, NOTICE_ID);
	SendRet (m->RetTid, m->RetPid, value);
	plist_del (object_plist, m->pid);
	free (m);
	}

int make_msgbox (Frame frame, int options, char *text, char *title, int Tid, int Pid)
	{
	struct MsgboxStruct *m;

	m = (struct MsgboxStruct *)malloc (sizeof (struct MsgboxStruct));
	if (m==NULL)
		SendRet (Tid, Pid, -1);
	m->magic = PID_MAGIC_MSGBOX;
	m->RetTid = Tid;
	m->RetPid = Pid;
	m->pid =  plist_add (object_plist, m);
	m->frame = frame;
	m->notice = (Xv_notice) xv_create ( frame, NOTICE,
		NOTICE_LOCK_SCREEN,	FALSE,
		NOTICE_BLOCK_THREAD,	FALSE,
		NOTICE_MESSAGE_STRING,	title,
		NOTICE_MESSAGE_STRING,	text,
		NOTICE_EVENT_PROC,	notice_event_proc,
		XV_KEY_DATA,	NOTICE_ID,	m,
		NULL);
	switch (options) {
		case MB_OK:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Ok", 	MB_ID_OK,	
				NULL);
				break;
		case MB_OKCANCEL:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Ok", 		MB_ID_OK,	
				NOTICE_BUTTON,	"Cancel",	MB_ID_CANCEL,
				NULL);
				break;
		case MB_RETRYCANCEL:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Retry",	MB_ID_RETRY,
				NOTICE_BUTTON,	"Cancel", 	MB_ID_CANCEL,	
				NULL);
				break;
		case MB_ABREIG:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Abort", 	MB_ID_ABORT,	
				NOTICE_BUTTON,	"Retry",	MB_ID_RETRY,
				NOTICE_BUTTON,	"Ignore",	MB_ID_IGNORE,
				NULL);
				break;
		case MB_YESNOCANCEL:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Yes", 		MB_ID_YES,	
				NOTICE_BUTTON,	"No", 		MB_ID_NO,	
				NOTICE_BUTTON,	"Cancel",	MB_ID_CANCEL,
				NULL);
				break;
		case MB_YESNO:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Yes", 		MB_ID_YES,	
				NOTICE_BUTTON,	"No",		MB_ID_NO,
				NULL);
				break;
		default:
			xv_set (m->notice, 
				NOTICE_BUTTON,	"Ok", 	MB_ID_OK,	
				NULL);
		}
	xv_set (m->notice, XV_SHOW,	TRUE, 	NULL);
	return m->pid;
	}

int destroy_msgbox (struct MsgboxStruct *m)
	{
	m->magic = 0;
	xv_destroy_safe (m->notice);
	free (m);
	}
