/***************************************************************
 *	
 *	object management for pidlist members.
 *
 */

#include  "xwio.h"
struct ObjectStruct {
	int magic;
	};

extern struct pidref *object_plist;


int destroy_object (int pid)
/*
 *	remove object
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, pid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				return destroy_swin ((struct SwinStruct *)plist_del (object_plist, pid));
			case PID_MAGIC_GWIN:
				return destroy_gwin ((struct GraphwinStruct *)plist_del (object_plist, pid));
			case PID_MAGIC_FILE:
				close_file ((struct FileStruct *)plist_del (object_plist, pid));
				return pid;
			case PID_MAGIC_MSGBOX:
				return destroy_msgbox ((struct MsgboxStruct *)plist_del (object_plist, pid));
			default:
				printf ("destroy object: illegal magic.\n");
			}
	else 
		printf ("destroy object: invalid id: %d.\n",pid);
	return -1;
	}

void write_object (int pid, char *str)
/*
 *	write string into object
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, pid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				simple_write (&(((struct SwinStruct *)obj)->twin), str);
				break;
			case PID_MAGIC_FILE:
				write_file ((struct FileStruct *)obj, str);
				break;
			default:
				printf ("write object: illegal magic (pid %d, str %s).\n", pid, str);
			}
	else 
		printf ("write object: invalid id: %d.\n",pid);
	}

void write_n_object (int pid, int n, char *str)
/*
 *	write n bytes into object...
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, pid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				simple_write_n (&(((struct SwinStruct *)obj)->twin), n, str);
				break;
	
			case PID_MAGIC_FILE:
				write_n_file ((struct FileStruct *)obj, n, str);
				break;
			default:
				printf ("write_n object: illegal magic (pid %d, str %s).\n", pid, str);
			}
	else 
		printf ("write_n object: invalid id: %d.\n",pid);
	}

void readln_object (int oid, int tid, int pid)
/*
 *	read line from object...
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, oid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				simple_readln (&(((struct SwinStruct *)obj)->twin), tid, pid);
				return;
			case PID_MAGIC_FILE:
				readln_file ((struct FileStruct *)obj, tid, pid);
				return;
			default:
				printf ("readln object: illegal magic (pid %d).\n", pid);
			}
	else 
		printf ("readln object: invalid id: %d.\n",oid);
	SendRet (tid, pid, -1);
	}

void readch_object (int oid, int tid, int pid)
/*
 *	read char from object...
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, oid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				simple_getch (&(((struct SwinStruct *)obj)->twin), tid, pid);
				return;
			case PID_MAGIC_FILE:
				readch_file ((struct FileStruct *)obj, tid, pid);
				return;
			default:
				printf ("readch object: illegal magic (pid %d).\n", pid);
			}
	else 
		printf ("readch object: invalid id: %d.\n",oid);
	SendRet (tid, pid, -1);
	}

void read_object (int oid, int n, int tid, int pid)
/*
 *	read n bytes from object...
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, oid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_FILE:
				read_file ((struct FileStruct *)obj, n, tid, pid);
				return;
			default:
				printf ("read object: illegal magic (pid %d).\n", pid);
			}
	else 
		printf ("read object: invalid id: %d.\n",oid);
	SendRet (tid, pid, -1);
	}

void clear_object (int pid)
/*
 *	
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, pid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				simple_clear (&(((struct SwinStruct *)obj)->twin));
				break;
			default:
				printf ("clear object: illegal magic (pid %d).\n", pid);
			}
	else 
		printf ("clear object: invalid id: %d.\n",pid);
	}

void text_object (int pid, char *str)
/*
 *	
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, pid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				xv_set (((struct SwinStruct *)obj)->frame, FRAME_LABEL, str, NULL);
				break;
			case PID_MAGIC_GWIN:
				xv_set (((struct GraphwinStruct *)obj)->frame, FRAME_LABEL, str, NULL);
				break;
			default:
				printf ("text object: illegal magic (pid %d, str %s).\n", pid, str);
			}
	else 
		printf ("text object: invalid id: %d.\n",pid);
	}

int object_getwin (int pid)
/*
 *	return Frame associated with window object...
 */
	{
	struct ObjectStruct *obj;

	obj=(struct ObjectStruct *)plist_get (object_plist, pid);
	if (obj)
		switch (obj->magic)
			{
			case PID_MAGIC_SWIN:
				return ((struct SwinStruct *)obj)->frame;
			case PID_MAGIC_GWIN:
				return ((struct GraphwinStruct *)obj)->frame;
			case PID_MAGIC_MSGBOX:
				return ((struct MsgboxStruct *)obj)->frame;
			default:
				printf ("object getwin: illegal magic (pid %d).\n", pid);
			}
	else 
		printf ("object getwin: invalid id: %d.\n",pid);
	return 0;
	}

