/*
	This Routines do the parsing of (flat) Prolog Terms into a
	printable string.
*/



/**************	Tags and Masks for Structures ***********************/

/* 1st Group: These are typed POINTERS: to extract the pointer value, just do "& MASK_TAG"  */

#define PTR_TAG		0x00	/* a pointer is recognizable because it points to a word address */
#define RSRVD_TAG	0x01	/* may later be used for extension */
#define LIST_TAG	0x02	/* points to a list - saves 4 words compared to V1.5 */
#define EXT2_TAG	0x03	/* element belongs to 2nd group */

/* 2nd Group: These are DATA elements with 4 Bit TAG: you still need a pointer to them */

#define ATOM_TAG	0x03	/* An atom needs information of length and hash anyway, so it can't be just a pointer */
#define INT_TAG		0x07	/* Isn't a pointer anyway, so why put into the upper group ? */
#define STRUCT_TAG	0x0B	/* a struct needs further information (ArcCount), so it can't be a pointer */
#define EXT3_TAG	0x0F	/* EXTENSION for the elements not belonging to this group */

/* 3rd Group: These are DATA elements with 8 Bit TAG */

#define NIL		0x0F  /* !!! this is a CONSTANT to indicate a free variable */
#define CHAR_TAG	0x1F	/* indicates a Character type, similar to an integer */
#define FUNC_TAG	0x2F	/* shall be used to manage predicates or functions */
#define EMPTY_LIST	0x3F  /* !!! CONSTANT: just an empty list - end of every regular list */

/**************	MASK_TAGS ***********************/

#define MASK_TAG	0xFFFFFFFC
#define MASK_TAG_SMALL	0xFFFFFFFC
#define MASK_TAG_PTR	0xFFFFFFFC
#define MASK_TAG_RSRVD	0xFFFFFFFC
#define MASK_TAG_LIST	0xFFFFFFFC

#define MASK_TAG_M	0xFFFFFFF0
#define MASK_TAG_MEDIUM	0xFFFFFFF0
#define MASK_TAG_ATOM	0xFFFFFFF0
#define MASK_TAG_INT	0xFFFFFFF0
#define MASK_TAG_STRUCT	0xFFFFFFF0

#define MASK_TAG_L	0xFFFFFF00
#define MASK_TAG_LARGE	0xFFFFFF00
#define MASK_TAG_CHAR	0xFFFFFF00
#define MASK_TAG_FUNC	0xFFFFFF00

/**************	TAG_MASKS ***********************/

#define TAG_MASK	0x00000003
#define TAG_MASK_SMALL	0x00000003
#define TAG_MASK_PTR	0x00000003
#define TAG_MASK_RSRVD	0x00000003
#define TAG_MASK_LIST	0x00000003

#define TAG_MASK_M	0x0000000F
#define TAG_MASK_MEDIUM	0x0000000F
#define TAG_MASK_ATOM	0x0000000F
#define TAG_MASK_INT	0x0000000F
#define TAG_MASK_STRUCT	0x0000000F

#define TAG_MASK_L	0x000000FF
#define TAG_MASK_LARGE	0x000000FF
#define TAG_MASK_CHAR	0x000000FF
#define TAG_MASK_FUNC	0x000000FF

/**************	SHIFT_VALUES ***********************/

#define TAG_LEN_PTR	0x02
#define TAG_LEN_RSRVD	0x02
#define TAG_LEN_LIST	0x02

#define TAG_LEN_ATOM	0x04
#define TAG_LEN_INT	0x04
#define TAG_LEN_STRUCT	0x04

#define TAG_LEN_CHAR	0x08
#define TAG_LEN_FUNC	0x08

/**************	HASHING ***********************/

#define HASH_MASK	0xFF000000
#define MASK_HASH	0x00FFFFFF


#define NIBBLE_MASK	0xF0000000	/* to check whether a value is data check */
#define DATA_TAG	0x40000000	/* for (X & NIBBLE_MASK) == DATA_TAG */

#include <stdlib.h>
#include <stdio.h>




int sPrintMsg(char *dest, int *dp, char *source, int ptr)
/* print prolog term into string.
   constructed from complie_msg_body () of dbserver.c   */

	{
	int	term;
	int	i;
	int	len;

	term = bufint (source+ptr);

	if ((term & TAG_MASK_LIST) == LIST_TAG)
		{
		len = (int)(term >> TAG_LEN_LIST);
		ptr +=4;
		sprintf (dest+*dp, "[");
		(*dp)++;
		for(i=0;i<len;i++)
			{
			ptr = sPrintMsg (dest, dp, source, ptr);
			if (i<len-1)
				{
				sprintf(dest+*dp, ",");
				(*dp)++;
				}
			}
		sprintf (dest+*dp, "]");
		(*dp)++;
		return(ptr);
		}

	if ((term & TAG_MASK_INT) == INT_TAG)
		{
		if (term<0)
			{
			term>>=4;
			term|=0xF0000000;
			}
		else
			term>>=4;

		sprintf(dest+*dp, "%ld", term);
		(*dp)+=strlen (dest+*dp);
		return(ptr+4);
		}

	if ((term & TAG_MASK_ATOM) == ATOM_TAG)
		{
		len = (int)((term & MASK_HASH) >> TAG_LEN_ATOM);
		ptr +=4;
		for(i=0;i<len*4&&source[ptr+i];i++)
			{
			if (source[ptr+i])
				{
				if (source[ptr+i]=='\n')
					sprintf (dest+*dp, "\\n");
				else
					sprintf (dest+*dp, "%c",source[ptr+i]);
				*dp+=strlen (dest+*dp);
				}
			}
		return(ptr+len*4);
		}

	if ((term & TAG_MASK_STRUCT) == STRUCT_TAG)
		{
		len = (int)(term >> TAG_LEN_STRUCT);
		ptr+=4;
		ptr = sPrintMsg (dest, dp, source, ptr);
		sprintf (dest+*dp, "(");
		(*dp)++;
		for(i=0; i<len; i++)
			{
			ptr = sPrintMsg (dest, dp, source, ptr);
			if (i<len-1)
				{
				sprintf(dest+*dp, ",");
				(*dp)++;
				}
			}
		sprintf (dest+*dp, ")");
		(*dp)++;
		return(ptr);
		}
	return (0);
	}

void PrintMsg (char *dest, char *source)
/* ... */
	{
	int dp=0;
	sPrintMsg (dest, &dp, source, 0);
	}
