#include "xwio.h"
#include <stdio.h>
#include <stdlib.h>


extern struct list *proc_list;
struct	{
	Frame	frame;
	Frame	parent;
	Panel_item	list;
	} proc_select;

struct ProcDataStruct *FindProcData (int Tid, int Pid);


void process_list_add (struct ProcDataStruct *p)
	{
	char	s[PROC_NAME_SIZE+3];

	sprintf (s, "%d/%d ",p->Tid,p->Pid);
	strncat (s, p->ProcName, PROC_NAME_SIZE-strlen (s));
	s[PROC_NAME_SIZE-1]='\0';
	xv_set (proc_select.list,
		PANEL_LIST_INSERT,	0,
		PANEL_LIST_STRING,	0,	s,
		PANEL_LIST_CLIENT_DATA, 0,	p,
		PANEL_LIST_DISPLAY_ROWS,	10,
		NULL);
	}

int process_list_get_row (struct ProcDataStruct *p)
	{
	int	i;
	struct ProcDataStruct *q;

        i=(int) xv_get (proc_select.list, PANEL_LIST_NROWS);

        while (i--)
                {
                q=(struct ProcDataStruct *)xv_get (proc_select.list, PANEL_LIST_CLIENT_DATA, i);
                if (q==p)
                        return i;
                }
        return -1;
        }

void process_list_set (struct ProcDataStruct *p)
	{
	int i;
	char	s[PROC_NAME_SIZE+3];

	if (!proc_select.frame)
		return;
	i=process_list_get_row (p);
	if (i != -1)
		{
		sprintf (s, "%d/%d ",p->Tid,p->Pid);
		strncat (s, p->ProcName, PROC_NAME_SIZE-strlen (s));
		s[PROC_NAME_SIZE-1]='\0';
		xv_set (proc_select.list, 
			PANEL_LIST_STRING,	i, s,
			PANEL_LIST_CLIENT_DATA,	i, p,
			NULL);
		}
	else
		process_list_add (p);
	}

void process_list_del (struct ProcDataStruct *p)
	{
	int i;

	if (!proc_select.frame)
		return;
	i=process_list_get_row (p);
	if (i != -1)
		xv_set (proc_select.list, PANEL_LIST_DELETE, i, NULL);
	}

#define P_TID	7
#define P_PID	8

void other_proc_ok (Panel_item item, Event *event)
/*
 *	"Ok" button notify proc,
 *	end the other dlg.
 */
	{
	int tid, pid;
	Panel_item ptid,ppid;

	ptid=(Panel_item) xv_get (item, XV_KEY_DATA, P_TID);
	ppid=(Panel_item) xv_get (item, XV_KEY_DATA, P_PID);
	tid=(int) xv_get (ptid, PANEL_VALUE);
	pid=(int) xv_get (ppid, PANEL_VALUE);
	xv_destroy_safe (xv_get (item, PANEL_CLIENT_DATA));
	if (tid)
		{
		int	msg[4];
		msg[0]=5;
		msg[1]=pid;
		msg[2]=1;
		SendSysMessage (tid, 0, 3, msg);
		TraceOn (tid, pid);
		}
	}

void other_proc_cancel (Panel_item item, Event *event)
/*
 *	"Cancel" button notify proc,
 *	end the other dlg.
 */
	{
	xv_destroy_safe ((Frame) xv_get (item, PANEL_CLIENT_DATA));
	}

void process_list_other (Panel_item item, Event *event)
/* 
 *	destroy the process list frame...
 */
	{
	Panel_item	ptid,ppid;
	Frame 	frame;
	Panel	panel;

	frame= (Frame) xv_create (proc_select.frame, FRAME,
		FRAME_LABEL,		"Add process",
		XV_WIDTH,		180,
		XV_HEIGHT,		80,
		NULL);
	panel= (Panel) xv_create (frame, PANEL, NULL); 
	ptid=(Panel_item) xv_create (panel, PANEL_NUMERIC_TEXT,
		PANEL_LABEL_STRING,	"Tid :",
		NULL);
	ppid=(Panel_item) xv_create (panel, PANEL_NUMERIC_TEXT,
		PANEL_NEXT_ROW,		-1,
		PANEL_LABEL_STRING,	"Pid :",
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_NEXT_ROW,		-1,
		PANEL_LABEL_STRING,	"Ok",
		PANEL_NOTIFY_PROC,	other_proc_ok,
		XV_KEY_DATA,	P_TID,	ptid,
		XV_KEY_DATA,	P_PID,	ppid,
		PANEL_CLIENT_DATA,	frame,
		NULL);
	xv_create (panel, PANEL_BUTTON,
		PANEL_LABEL_STRING,	"Cancel",
		PANEL_NOTIFY_PROC,	other_proc_cancel,
		PANEL_CLIENT_DATA,	frame,
		NULL);
	window_fit (frame);
	xv_set (frame, XV_SHOW, TRUE, NULL);
	}

void process_list_ok (Panel_item item, Event *event)
/* 
 *	add the selected process to the trace list.
 */
	{
	int	s;
	struct ProcDataStruct *p;

	s=(int) xv_get (proc_select.list, PANEL_LIST_FIRST_SELECTED);
	if (s!=-1)
		{
		int	msg[4];
		p=(struct ProcDataStruct *)xv_get (proc_select.list, PANEL_LIST_CLIENT_DATA, s);
		msg[0]=5;
		msg[1]=p->Pid;
		msg[2]=1;
		SendSysMessage (p->Tid, 0, 3, msg);
                TraceOn (p->Tid, p->Pid);
		}
	xv_destroy_safe (proc_select.frame);
	proc_select.frame=0;
	}

void process_list_cancel (Panel_item item, Event *event)
/* 
 *	destroy the process list frame...
*/
	{
	xv_destroy_safe (proc_select.frame);
	proc_select.frame=0;
	}

void process_list (Frame parentframe, int addflag)
/*
 *	display the process list select dialog...
 */

	{
	struct list *elem;
	struct ProcDataStruct *p;
	Panel 	panel;

	if (proc_select.frame)
		{
		if (proc_select.parent==parentframe)
			{
			xv_destroy_safe (proc_select.frame);
			proc_select.frame=0;
			}
		return;
		}
	proc_select.parent=parentframe;
	proc_select.frame= (Frame) xv_create (parentframe, FRAME,
		FRAME_LABEL,		"Process list",
		WIN_COLUMNS,		40,
		WIN_ROWS,		19,
		NULL);
	panel= (Panel)  xv_create (proc_select.frame, PANEL, NULL);
	if (addflag)
		{
		xv_create (panel, PANEL_BUTTON,
			PANEL_NOTIFY_PROC,	process_list_ok,
			PANEL_LABEL_STRING,	"Ok",
			NULL);
		xv_create (panel, PANEL_BUTTON,
			PANEL_NOTIFY_PROC,	process_list_cancel,
			PANEL_LABEL_STRING,	"Cancel",
			NULL);
		xv_create (panel, PANEL_BUTTON,
			PANEL_NOTIFY_PROC,	process_list_other,
			PANEL_LABEL_STRING,	"Other",
			NULL);
		}
	else
		xv_create (panel, PANEL_BUTTON,
			PANEL_NOTIFY_PROC,	process_list_cancel,
			PANEL_LABEL_STRING,	"Ok",
			NULL);
	proc_select.list=(Panel_item)xv_create (panel, PANEL_LIST,
		PANEL_NEXT_ROW,		-1,
		PANEL_LIST_WIDTH,	-1,
		PANEL_LIST_DISPLAY_ROWS,	10,
		NULL);
	elem=get_first_list_elem (proc_list);
	while (elem)
		{
		p=(struct ProcDataStruct *)elem->data;
		process_list_add (p);
		elem=get_next_list_elem (elem);
		}
	window_fit (proc_select.frame);
	xv_set (proc_select.frame, XV_SHOW, TRUE, NULL);
	}

int AddProcess (int Tid, int Pid, char *ProcName)
/* Add new Process to List */
	{
	struct ProcDataStruct *p;
	struct list *elem;
	int	i;

	if (!(p=FindProcData (Tid, Pid)))
		{
		elem=(struct list *)malloc (sizeof (struct list));
		p=(struct ProcDataStruct *)malloc (sizeof (struct ProcDataStruct));
		p->Tid=Tid;
		p->Pid=Pid;
		elem->data=p;
		assertz_list (proc_list, elem);
		}
	strncpy (p->ProcName, ProcName, PROC_NAME_SIZE-1);
	p->ProcName[PROC_NAME_SIZE-1]='\0';

	SetProcName (Tid, Pid, ProcName);

	process_list_set (p);
	return 1;
	}


int DelProcess (int Tid, int Pid)
/* Del Process from List */
	{
	struct ProcDataStruct *p;
	struct list *elem;

	elem=get_first_list_elem (proc_list);
	while (elem)
		{
		p=(struct ProcDataStruct *)elem->data;
		if (p->Tid==Tid && p->Pid==Pid)
			{
			process_list_del (p);
			remove_list_elem (proc_list, elem);
			free (elem);
			free (p);
			return TRUE; 
			}
		elem=get_next_list_elem (elem);
		}
	return FALSE;
	}

void ClearProcess ()
/*
 *	Empty processlist
 */
	{
	struct list *elem;
	elem=get_first_list_elem (proc_list);
	while (elem)
		{
		remove_list_elem (proc_list, elem);
		free (elem->data);
		free (elem);
		elem=get_first_list_elem (proc_list);
		}
	}

struct ProcDataStruct *FindProcData (int Tid, int Pid)
/* returns Pointer to ProcessDataStruct */
	{
	struct list *elem;
	struct ProcDataStruct *p;

	elem=get_first_list_elem (proc_list);
	while (elem != NULL)
		{
		p=(struct ProcDataStruct *)elem->data;
		if (p->Tid==Tid && p->Pid==Pid)
			return p;  /* Process found */
		elem=get_next_list_elem (elem);
		}
	return NULL;
	}

