
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <xview/xview.h>
#include <xview/canvas.h>
#include <xview/xv_xrect.h>
#include <xview/scrollbar.h>
#include <stdlib.h>
#include <memory.h>

#include "xwio.h"

void simple_refresh (struct TwinStruct *twin);
void simple_line (struct TwinStruct *twin);
void	twin_event_proc (), twin_repaint_proc ();
void read_char (struct TwinStruct *twin);
void simple_write_char (struct TwinStruct *twin, char s, int drawflag);

extern struct pidref *object_plist;
extern	GC	gc;
extern	XFontStruct *font;
extern struct ConsoleStruct console;
extern Menu twin_menu;



int make_twin_sub (Frame frame, int rows, int cols, struct TwinStruct *twin)
/*
 *	create a twin text window to fit into the frame...
 */
	{
	int	charwidth, charheight;
	if (font == NULL) {
		charwidth = 6;
		charheight = 13;
		}
	else 	{
		charwidth = XTextWidth (font, " ", 1);
		charheight = font->max_bounds.ascent+font->max_bounds.descent;
		}
	twin->canvas= (Canvas) xv_create (frame, CANVAS,
		/*
		WIN_COLUMNS,		cols+2,
		WIN_ROWS,		rows+2,
		*/
		XV_WIDTH,		(cols +2) * charwidth + 40,
		XV_HEIGHT,		(rows +2) * charheight + 15,
		CANVAS_X_PAINT_WINDOW,	TRUE,
		CANVAS_REPAINT_PROC,	twin_repaint_proc,
		CANVAS_AUTO_SHRINK,	FALSE,
		CANVAS_AUTO_EXPAND,	FALSE,
		XV_KEY_DATA,	TWIN_ID,	twin,
		NULL);
	twin->scrollbar_h= (Scrollbar) xv_create (twin->canvas, SCROLLBAR,
		SCROLLBAR_DIRECTION,	SCROLLBAR_HORIZONTAL,
		SCROLLBAR_SPLITTABLE,	FALSE,
		NULL);
	twin->scrollbar_v= (Scrollbar) xv_create (twin->canvas, SCROLLBAR,
		SCROLLBAR_DIRECTION,	SCROLLBAR_VERTICAL,
		SCROLLBAR_SPLITTABLE,	FALSE,
		NULL);
		
	xv_set (canvas_paint_window (twin->canvas), 
		WIN_EVENT_PROC,		twin_event_proc,
		WIN_CONSUME_EVENTS,
			WIN_ASCII_EVENTS,
			WIN_MOUSE_BUTTONS,
			NULL,
		XV_KEY_DATA,	TWIN_ID,	twin,
		NULL);

	twin->size_x=cols;
	twin->size_y=rows;
	twin->start_y=0;
	twin->cur_x=0;
	twin->cur_y=0;
	twin->screen=(char *) malloc ((cols+1)*(rows+1));
	if (!twin->screen)
		fprintf (stderr, "twin unable to alloc screen mem.\n");
	memset (twin->screen, ' ', cols*rows);
	twin->char_count=0;
	twin->char_next=0;
	twin->read_flag=0;
	twin->in_buf[0]=0;
	return TRUE;
	}

void destroy_twin_sub (struct TwinStruct *twin)
/*
 *	Kill twin window...
 */
	{
	free (twin->screen);
	xv_destroy_safe (twin->canvas);
	}

void do_char (Xv_window window, char ch)
	{
	struct TwinStruct *twin;

	twin= (struct TwinStruct *) xv_get (window, XV_KEY_DATA, TWIN_ID);
	if (twin->char_count<CHAR_BUFFER_SIZE)
		if (twin->char_next+twin->char_count < CHAR_BUFFER_SIZE)
			twin->char_buf[twin->char_next+(twin->char_count++)]=ch;
		else
			twin->char_buf[twin->char_next+(twin->char_count++)-CHAR_BUFFER_SIZE]=ch;
	/* else beep */
	switch (twin->read_flag){
		case IN_READLINE:
			read_char (twin);
			break;

		case IN_GETCH:
			{
			char ss[2];

			SendRet (twin->in_tid, twin->in_pid, (long)twin->char_buf[twin->char_next]);
			twin->read_flag=0;
/*  use this for	ss[0]=twin->char_buf[twin->char_next];
 *  local echo.		ss[1]=0;
 *			simple_write (twin, ss);
 */
			twin->char_count--;
			if (twin->char_next<CHAR_BUFFER_SIZE-1)
				twin->char_next++;
			else
				twin->char_next=0;
			break;
			}
		}
	}

void twin_event_proc (Xv_window window, Event *event)
	{
	if (event_is_ascii (event) && event_is_down (event))
		do_char (window, event_action (event));
	else	{
		switch (event_action (event))
			{
			case ACTION_MENU:
				menu_show (twin_menu, window, event, NULL);
				break;
			case ACTION_ERASE_CHAR_BACKWARD:
				if (event_is_down (event))
					do_char (window, 0x08);
				break;
			}
		}
	return;
	}

void twin_repaint_proc (Canvas canvas, Xv_window paint_window, Display *dpy, Window xwin, Xv_xrectlist *xrects)
	{
	struct TwinStruct *twin;
	int	font_height;
	int	i,j,w, r;
	Rect	*rect;

	Region	region;
	XRectangle  rectangle;

	twin= (struct TwinStruct *)xv_get (paint_window, XV_KEY_DATA, TWIN_ID);

	region = XCreateRegion ();
	if (xrects == NULL)
		{
		rectangle.x = 0;
		rectangle.y = 0;
		rectangle.width = xv_get (paint_window, XV_WIDTH);
		rectangle.height = xv_get (paint_window, XV_HEIGHT);
		XUnionRectWithRegion (&rectangle, region, region);
		}
	else
		for (i=0; i < xrects->count; i++)	
			XUnionRectWithRegion (&(xrects->rect_array[i]), region, region);

	/*
	XSetForeground (console.dpy, gc, BlackPixel (console.dpy, console.screen_no));
	XSetBackground (console.dpy, gc, WhitePixel (console.dpy, console.screen_no));
	*/


	/*XClearWindow (dpy, xwin); */

	font_height=font->max_bounds.ascent+font->max_bounds.descent;
	for (i=0; i<twin->size_y; i++)
		{
		j=twin->start_y+i;
		if (j>=twin->size_y)
			j-=twin->size_y;
		rectangle.x = 10;
		rectangle.y = 20+(i*font_height);
		rectangle.width = XTextWidth (font, twin->screen+j*twin->size_x, twin->size_x);
		rectangle.height = font_height;
		r = XRectInRegion (region, 10, 20+(i*font_height),
					XTextWidth (font, twin->screen+j*twin->size_x, twin->size_x),
					font_height);
		if (r == RectangleIn || r == RectanglePart )
			{
			XClearArea (dpy, xwin, 10, 20+(i*font_height)-font->max_bounds.ascent, 0, font_height, False);
			XDrawString (dpy, xwin, gc, 10, 20+(i*font_height), twin->screen+j*twin->size_x, twin->size_x);
			}
		}
	if (twin->read_flag)
		{
		i=twin->cur_y-twin->start_y;
		if (i<0)
			i+=twin->size_y;

		w=XTextWidth (font, twin->screen+j*twin->size_x, twin->cur_x);
		XDrawString (dpy, xwin, gc, 10+w, 20+(i*font_height), "_", 1);
		}
	XDestroyRegion (region);
	}

void scrollup (struct TwinStruct *twin)
	{
	Display * dpy;
	Window xwin;
	int	font_height;

	dpy = (Display *) xv_get (twin->canvas, XV_DISPLAY);
	xwin = xv_get (canvas_paint_window (twin->canvas), XV_XID);

	font_height=font->max_bounds.ascent+font->max_bounds.descent;
	XCopyArea (dpy, xwin, xwin, gc, 
		10, 20 - font->max_bounds.ascent + font_height,
		xv_get (canvas_paint_window (twin->canvas), XV_WIDTH),
		xv_get (canvas_paint_window (twin->canvas), XV_HEIGHT)- (20 - font->max_bounds.ascent),
		10, 20 - font->max_bounds.ascent);
	}

	
void simple_line (struct TwinStruct *twin)
	{
	int	font_height;
	int	j,w;
	Display *dpy;
	Xv_window	xwin;

	/*
	XSetForeground (console.dpy, gc, BlackPixel (console.dpy, console.screen_no));
	XSetBackground (console.dpy, gc, WhitePixel (console.dpy, console.screen_no));
	*/

	xwin=xv_get (canvas_paint_window (twin->canvas), XV_XID);
	dpy= (Display *)xv_get (twin->canvas, XV_DISPLAY);

	font_height=font->max_bounds.ascent+font->max_bounds.descent;
	j=twin->cur_y-twin->start_y;
	if (j<0)
		j+=twin->size_y;

	w=XTextWidth (font, twin->screen+twin->cur_y*twin->size_x, twin->size_x);
	XClearArea (dpy, xwin, 10+w, 20+(j*font_height)-font->max_bounds.ascent, 0, font_height, False);
	XDrawImageString (dpy, xwin, gc, 10, 20+(j*font_height), twin->screen+twin->cur_y*twin->size_x, twin->size_x);
	if (twin->read_flag)
		{
		w=XTextWidth (font, twin->screen+twin->cur_y*twin->size_x, twin->cur_x);
		XDrawString (dpy, xwin, gc, 10+w, 20+(j*font_height), "_", 1);
		}
	}

void simple_char (struct TwinStruct *twin)
	{
	simple_line (twin);
	return;
	}

void simple_refresh (struct TwinStruct *twin)
	{
	twin_repaint_proc (twin->canvas, canvas_paint_window (twin->canvas), (Display *)xv_get (twin->canvas,XV_DISPLAY), xv_get (canvas_paint_window (twin->canvas), XV_XID), (Xv_xrectlist *)NULL);
	}

void read_char (struct TwinStruct *twin)
/* Process keyboard input for readln */
	{
	if (!twin->char_count)
		simple_line (twin);	/* display text cursor... */
	while (twin->char_count && twin->char_buf[twin->char_next]!='\r' && twin->char_buf[twin->char_next]!=27)
		{
		if (twin->char_buf[twin->char_next]==8)
			{
			if (strlen (twin->in_buf))
				{
				simple_write (twin, "\x08" );
				twin->in_buf[strlen(twin->in_buf)-1]=0;
				}
			}
		else
			if (strlen (twin->in_buf)<READ_BUFFER_SIZE)
				{
				twin->in_buf[strlen (twin->in_buf)+1]=0;
				twin->in_buf[strlen (twin->in_buf)]=twin->char_buf[twin->char_next];
				simple_write_char (twin, twin->in_buf[strlen(twin->in_buf)-1], 1);
				}
		twin->char_count--;
		if (twin->char_next<CHAR_BUFFER_SIZE-1)
			twin->char_next++;
		else
			twin->char_next=0;
		}
	if (twin->char_count && (twin->char_buf[twin->char_next]=='\r' || twin->char_buf[twin->char_next]==27))
		{
		twin->read_flag=0;
		simple_write (twin, "\n");
		if (twin->char_buf[twin->char_next]==27)
			SendRet (twin->in_tid, twin->in_pid, -1L);
		else
			SendRetStr (twin->in_tid, twin->in_pid, twin->in_buf);
		twin->char_count--;
		if (twin->char_next<CHAR_BUFFER_SIZE-1)
			twin->char_next++;
		else
			twin->char_next=0;
		twin->in_buf[0]=0;
		}
	}

void simple_readln (struct TwinStruct *twin, int tid, int pid)
/* switch window into readln mode, enable cursor, position window... */
	{
	if (twin->read_flag)
		return ;
	twin->read_flag=IN_READLINE;
	twin->in_pid=pid;
	twin->in_tid=tid;
	twin->in_buf[0]='\0';
	read_char (twin);
	}

void simple_getch (struct TwinStruct *twin, int tid, int pid)
	{
	if (twin->read_flag)
		return ;
	if (twin->char_count)
		{
		SendRet (tid, pid, (long)twin->char_buf[twin->char_next]);
		twin->read_flag=IN_GETCH;

/*		simple_write_char (twin, twin->char_buf[twin->char_next], 1);
 */
		twin->read_flag=0;
		twin->char_count--;
		if (twin->char_next<CHAR_BUFFER_SIZE-1)
			twin->char_next++;
		else
			twin->char_next=0;
		}
	else
		{
		twin->read_flag=IN_GETCH;
		twin->in_pid=pid;
		twin->in_tid=tid;
		}
	return ;
	}

void simple_scroll (struct TwinStruct *twin)
/* Scroll visible area of SimpleWindow upwards one line */
	{
	simple_line (twin);
	twin->cur_x=0;
	twin->cur_y++;
	if(twin->cur_y>=twin->size_y)
		twin->cur_y=0;
	if(twin->cur_y==twin->start_y)
		{
		memset(&twin->screen[twin->start_y*twin->size_x],' ',twin->size_x);
		twin->start_y++;
		if(twin->start_y==twin->size_y)
			twin->start_y=0;
		scrollup (twin);
/*		simple_refresh (twin); */
		}
	}

void simple_write_char (struct TwinStruct *twin, char s, int drawflag)
/* Write char into Screen Buffer of Window hWnd */
	{
	#define TABCOLS	3

	switch(s)
		{
		case 8:
			if (twin->cur_x || twin->cur_y!=twin->start_y)
				{
				if (twin->cur_x==0)
					{
					simple_line (twin);
					twin->cur_x=twin->size_x;
					if (twin->cur_y)
						twin->cur_y--;
					else
						twin->cur_y=twin->size_y-1;
					}
				twin->cur_x--;
				twin->screen[twin->cur_y*twin->size_x+twin->cur_x]=' ';
				}
			break;
		case 9:
			twin->cur_x+=TABCOLS-(twin->cur_x % TABCOLS);
			if(twin->cur_x==twin->size_x)
				simple_scroll (twin);
			break;
		case 13:
			twin->cur_x=0;
			break;
		case 10:
			simple_scroll (twin);
			break;
		default:
				twin->screen[twin->cur_y*twin->size_x+twin->cur_x++]=s;
			if(twin->cur_x==twin->size_x)
				simple_scroll (twin);
		}
	if (drawflag)
		simple_char (twin);
	}

void simple_write_n (struct TwinStruct *twin, int i, char *str)
/* Write str into Screen Buffer of Window hWnd */
	{
	#define TABCOLS	3

	while (i--)
		if (str[i])
			simple_write_char (twin, str[i], 0);
	simple_line (twin);
	}

void simple_write(struct TwinStruct *twin, char *str)
/* Write str into Screen Buffer of Window hWnd */
	{
	int i=0;
	#define TABCOLS	3

	while (str[i]!=0)
		simple_write_char (twin, str[i++], 0);
	simple_line (twin);
	}

void simple_write_pid (int pid, char *str)
/* Write str into Screen Buffer of pid */
	{
	struct SwinStruct 	*swin;

	swin=plist_get (object_plist, pid);
	if (swin->magic != PID_MAGIC_SWIN)
		{
		printf ("wrong id %d for simple_write %s.\n",pid,str);
		return;
		}
	simple_write (&(swin->twin),str);
	}

void simple_clear (struct TwinStruct *twin)
/* ClearWindow */
	{
	memset (twin->screen, ' ', twin->size_x*twin->size_y);
	twin->start_y=0;
	twin->cur_x=0;
	twin->cur_y=0;
	simple_refresh (twin);
	}

void simple_clear_pid (int	pid)
/* ClearWindow */
	{
	struct	SwinStruct *swin;

	swin=plist_get (object_plist, pid);
	if (swin->magic != PID_MAGIC_SWIN)
		{
		printf ("wrong id %d for simple_clear.\n",pid);
		return;
		}
	simple_clear ( &(swin->twin));
	}

void simple_cursor (int pid, long x, long y)
/* 
 *	Set Cursor Position
 */
	{
	struct TwinStruct *twin;
	struct SwinStruct	*swin;
	
	swin=plist_get (object_plist, pid);
	if (swin->magic != PID_MAGIC_SWIN)
		{
		printf ("wrong id %d for simple_cursor.\n",pid);
		return;
		}
	twin=&(swin->twin);
	twin->cur_x=(int)x;
	twin->cur_y=(int)y;
}
long simple_getcolumn (int pid)
/* */
	{
	struct TwinStruct *twin;
	struct SwinStruct	*swin;
	
	swin=plist_get (object_plist, pid);
	if (swin->magic != PID_MAGIC_SWIN)
		{
		printf ("wrong id %d for simple_getcol.\n",pid);
		return;
		}
	twin=&(swin->twin);
	return twin->cur_x;
	}

long simple_getrow (int pid)
/* */
	{
	struct TwinStruct *twin;
	struct SwinStruct	*swin;
	
	swin=plist_get (object_plist, pid);
	if (swin->magic != PID_MAGIC_SWIN)
		{
		printf ("wrong id %d for simple_getrow.\n",pid);
		return;
		}
	twin=&(swin->twin);
	return twin->cur_y;
	}

