/* 		XWIO.H
 *		definitions include file...
 *	should be included everywhere.
 */

#include <errno.h>
/* #include <curses.h> */
#ifdef WINDOW
#undef WINDOW /* defined by curses.h -- needs to be undefined */
#endif
#include <xview/xview.h>
#include <xview/panel.h>
#include <xview/textsw.h>
#include <xview/tty.h>
#include <xview/scrollbar.h>
#include <xview/notice.h>
#ifdef USE_FILECHOOSER
#include <xview/file_chsr.h>
#endif


/*
 *	system build options
 */
#ifdef BIG_END
#define swint(A)	(A)
#define swi16(A)      (A)
#define swlist(A,B,C,D)	(D),(C),(B),(A)
#else
#define swlist(A,B,C,D)	(A),(B),(C),(D) 
#endif

#define noUSE_SOCKET_PSERVER	1
#define noUSE_LOCAL_PSERVER	1

#define noSPY_WINDOW

#define CHAR_BUFFER_SIZE	100
#define READ_BUFFER_SIZE	1000
#define IN_READLINE		1
#define IN_GETCH		2

#define TOUT_A		50

#ifndef HAS_STRERROR
extern char *sys_errlist[];
#define strerror(A)	sys_errlist[(A)]
#endif

struct TtyStruct {
        Tty     ttysw;
	int	pipe[2][2];
	int	pid;
        };

struct TwinStruct {
	Canvas	canvas;
	Scrollbar	scrollbar_h;
	Scrollbar	scrollbar_v;
	int	parent_pid;
	int	size_x;
	int	size_y;
	char	*screen;
	int	cur_x;
	int	cur_y;
	int	start_y;
	int	read_flag;
	char	char_buf[CHAR_BUFFER_SIZE];
	int	char_next;
	int	char_count;
	char	in_buf[READ_BUFFER_SIZE];
	int	in_pid;
	int	in_tid;
	};

struct SwinStruct {
	int	magic;
	Frame	frame;
	struct TwinStruct twin;
	};

struct GraphwinStruct {
        int     magic;
        Frame   frame;
        struct GwinStruct *gwin;
        };

struct MsgboxStruct {
	int	magic;
	Xv_notice	notice;
	int	pid;
	Frame	frame;
	int 	RetTid;
	int 	RetPid;
	};

#define	OM_READ		0
#define OM_WRITE	1
#define OM_APPEND	2
#define OM_MODIFY	3

struct FileStruct {
	int	magic;
	FILE	*fd;
	};

struct LinkStruct {
	int     id;             /* Link ID for link routines */
	char    name[16];       /* Link Driver Name */
	int     pipe[2][2];     /* communication with child  */
	int     in_pid;         /* input child process */
	int     out_pid;        /* output child */
	struct list *qOut;	/* Q for output blocks. */
	int	fOutReady;	/* flag for waiting out link */
	int	fPause;		
	};

#ifndef USE_FILECHOOSER
typedef int File_chooser;
#endif

struct ConsoleStruct {
	Frame		frame;
	Panel		panel;
	Tty		tty;
	struct TwinStruct	*twin;
	Panel_item	message;
	struct LinkStruct	*link;
	Display		*dpy;
	int		screen_no;
	File_chooser	openchooser;
	};

struct SpyStruct {
	Frame	frame;
	struct TwinStruct *twin;
	FILE	*f;
	};

#define SETUP_FILE_LEN	100
#define PROC_NAME_SIZE	100

struct SetupDataStruct {
        char    networkfile[SETUP_FILE_LEN];
        char    loadfile[SETUP_FILE_LEN];
        char    linkdriver[SETUP_FILE_LEN];
	char	helpviewer[256];
        int     spyonbootflag;
	int	spyfile;
        };


struct SetupStruct {
        Frame           frame;
        Panel_item      networkfile;
        Panel_item      loadfile;
        Panel_item      linkdriver;
        Panel_item      helpviewer;
        File_chooser    network_chsr;
        File_chooser    loadfile_chsr;
        struct SetupDataStruct  oldsetup;
        } ;

struct ProcDataStruct {
	int	Tid;
	int	Pid;
	char	ProcName[PROC_NAME_SIZE];
        };

struct list
	{
        struct list *next;
        struct list *prev;
        void        *data;
	};

#define PID_MAGIC_TTY		42
#define PID_MAGIC_SWIN		44
#define PID_MAGIC_FILE		46
#define PID_MAGIC_GWIN		48
#define PID_MAGIC_MSGBOX	50
#define HEADER_LEN		8
#define BLOCK_SIZE		512
#define TWIN_ID			70
#define DEBUG_ID		72
#define GWIN_ID			74
#define NOTICE_ID		76
#define SWIN_ID			77
#define	CHSR_KEY		78
#define SYSTEM_ID		79

#define KEY_SETUP_NETWORK	1
#define	KEY_SETUP_LOADFILE	2

#define T_NAMESIZE      50
#define T_DISPLAY_X     80
#define T_DISPLAY_Y     20

struct pidref {
        int     pid;
        void *data;
        };

#define max(a,b)            (((a) > (b)) ? (a) : (b))
#define min(a,b)            (((a) < (b)) ? (a) : (b))


/* FUNCTION PROTOTYPES.... */

 /* main.c */

void ERROR (char *str);
void error_box (char *str);
char *binfile (char *file) ;
char *find_file (char *file);
char *strlwr (char *str);

 /* open.c */

void open_lkd_dlg (Frame parent);

 /* tty.c */

int DestroyTty (struct SwinStruct *swin);
int make_swin (int startrow, int startcol, int rows, int cols, char *title);
void do_tty (struct TtyStruct *tty);
void make_tty_sub (struct TtyStruct *tty, Frame frame);
/* void tprintf (...); */

 /* twin.c */

void simple_write(struct TwinStruct *twin, char *str);
void simple_write_n (struct TwinStruct *twin, int n, char *str);
void simple_write_pid (int pid, char *str);
void simple_clear (struct TwinStruct *twin);
void simple_clear_pid (int      pid);
void simple_cursor (int pid, long x, long y);
void simple_readln (struct TwinStruct *twin, int tid, int pid);
int make_twin_sub (Frame frame, int rows, int cols, struct TwinStruct *twin);
void destroy_twin_sub (struct TwinStruct *twin);
long simple_getcolumn (int pid);
long simple_getrow (int pid);

 /*plist.c */

int plist_add();
struct pidref *plist_init ();
void *plist_get ();
void *plist_del();

 /* list.c */

struct list *init_list ();
int is_empty_list ();
struct list *get_first_list_elem ();
struct list *get_next_list_elem ();
void asserta_list ();
void assertz_list ();
void remove_list_elem ();
struct list *find_list_id ();
struct list *find_list_ptr ();
int list_member ();
int count_list_elems ();

 /* dolink.c */

void Write_Message (struct LinkStruct *link, int tid, int pid, char *buf, int len);
void file_str (struct LinkStruct *link, int Tid, int Pid, char *filename);
struct LinkStruct *InitMyLink (char *link_name);
int StartLinkLoops (struct LinkStruct *link);
void KillLink (struct LinkStruct *link);

 /* linkprim.c */

int ReadChunk ( int LinkId, char *Buffer, int len);
int WriteChunk ( int LinkId, char *Buffer, int len);
int ReadBlock (int LinkId, char *header, char *buffer);
int WriteBlock ( int LinkId, char *header, char *buffer);

 /* exprolog.c */

#ifndef swint
int swint (int i);
short swi16 (short i);
#endif
void intbuf (int l, char *buf);
void putint (int i, char *buf);
int bufint (char *buf);
int getint (char *buf);
char *SkipString (char *atom);
void SendSysMessage (int Tid, int Pid, int words, int *msg);
void SendRetStruct (int Tid, int Pid, char *sArg, int len);
void SendRetStr (int Tid, int Pid, char *szArg);
void SendRet (int Tid, int Pid, int arg);
void ExecPrologMessage (char *buf, int len);
void RunApp (char *File);

 /* spy.c */

void WriteMsgSpyIn (char *buf);
void WriteMsgSpyOut (char *Header, char *msg);

 /* parse.c */

void PrintMsg (char *dest, char *source);

 /* boot.c */

int Reset (struct LinkStruct *link);
void Worm (struct LinkStruct *link);

 /* objects.c */

int destroy_object (int pid);;
void write_object (int pid, char *str);
void clear_object (int pid);
void text_object (int pid, char *str);
int object_getwin (int pid);

 /* process.c */

void process_list (Frame parentframe, int addflag);
int AddProcess (int Tid, int Pid, char *ProcName);
int DelProcess (int Tid, int Pid);
struct ProcDataStruct *FindProcData (int Tid, int Pid);

 /* debug.c */

void trace_window (Frame parentframe);
int TraceOn (int Tid, long Pid);
int TraceOff (int Tid, int Pid);
int Trace (int RTid, int RPid, int Tid, int Pid, int FilePos, char *FileName, char *Text);
int TracePrint (int Tid, int Pid, char *Text);
void SetProcName (int Tid, long Pid, char *ProcName);

 /* graph.c */

int make_gwin (int x, int y, char *title);
int destroy_gwin (struct GraphwinStruct *g);
int DrawPolygon (int oid, char *buf);
int DrawEllipse (int oid, char *buf);
int EndDraw (int oid);

 /* msgbox.c */

int make_msgbox (Frame frame, int options, char *text, char *title, int Tid, int Pid);
int dextroy_msgbox (struct MsgboxStruct *m);
